# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import argparse
import sys

import six
from six.moves import zip

import charmhelpers.core.unitdata


class OutputFormatter(object):
    def __init__(self, outfile=sys.stdout):
        self.formats = (
            "raw",
            "json",
            "py",
            "yaml",
            "csv",
            "tab",
        )
        self.outfile = outfile

    def add_arguments(self, argument_parser):
        formatgroup = argument_parser.add_mutually_exclusive_group()
        choices = self.supported_formats
        formatgroup.add_argument("--format", metavar='FMT',
                                 help="Select output format for returned data, "
                                      "where FMT is one of: {}".format(choices),
                                 choices=choices, default='raw')
        for fmt in self.formats:
            fmtfunc = getattr(self, fmt)
            formatgroup.add_argument("-{}".format(fmt[0]),
                                     "--{}".format(fmt), action='store_const',
                                     const=fmt, dest='format',
                                     help=fmtfunc.__doc__)

    @property
    def supported_formats(self):
        return self.formats

    def raw(self, output):
        """Output data as raw string (default)"""
        if isinstance(output, (list, tuple)):
            output = '\n'.join(map(str, output))
        self.outfile.write(str(output))

    def py(self, output):
        """Output data as a nicely-formatted python data structure"""
        import pprint
        pprint.pprint(output, stream=self.outfile)

    def json(self, output):
        """Output data in JSON format"""
        import json
        json.dump(output, self.outfile)

    def yaml(self, output):
        """Output data in YAML format"""
        import yaml
        yaml.safe_dump(output, self.outfile)

    def csv(self, output):
        """Output data as excel-compatible CSV"""
        import csv
        csvwriter = csv.writer(self.outfile)
        csvwriter.writerows(output)

    def tab(self, output):
        """Output data in excel-compatible tab-delimited format"""
        import csv
        csvwriter = csv.writer(self.outfile, dialect=csv.excel_tab)
        csvwriter.writerows(output)

    def format_output(self, output, fmt='raw'):
        fmtfunc = getattr(self, fmt)
        fmtfunc(output)


class CommandLine(object):
    argument_parser = None
    subparsers = None
    formatter = None
    exit_code = 0

    def __init__(self):
        if not self.argument_parser:
            self.argument_parser = argparse.ArgumentParser(description='Perform common charm tasks')
        if not self.formatter:
            self.formatter = OutputFormatter()
            self.formatter.add_arguments(self.argument_parser)
        if not self.subparsers:
            self.subparsers = self.argument_parser.add_subparsers(help='Commands')

    def subcommand(self, command_name=None):
        """
        Decorate a function as a subcommand. Use its arguments as the
        command-line arguments"""
        def wrapper(decorated):
            cmd_name = command_name or decorated.__name__
            subparser = self.subparsers.add_parser(cmd_name,
                                                   description=decorated.__doc__)
            for args, kwargs in describe_arguments(decorated):
                subparser.add_argument(*args, **kwargs)
            subparser.set_defaults(func=decorated)
            return decorated
        return wrapper

    def test_command(self, decorated):
        """
        Subcommand is a boolean test function, so bool return values should be
        converted to a 0/1 exit code.
        """
        decorated._cli_test_command = True
        return decorated

    def no_output(self, decorated):
        """
        Subcommand is not expected to return a value, so don't print a spurious None.
        """
        decorated._cli_no_output = True
        return decorated

    def subcommand_builder(self, command_name, description=None):
        """
        Decorate a function that builds a subcommand. Builders should accept a
        single argument (the subparser instance) and return the function to be
        run as the command."""
        def wrapper(decorated):
            subparser = self.subparsers.add_parser(command_name)
            func = decorated(subparser)
            subparser.set_defaults(func=func)
            subparser.description = description or func.__doc__
        return wrapper

    def run(self):
        "Run cli, processing arguments and executing subcommands."
        arguments = self.argument_parser.parse_args()
        if six.PY2:
            argspec = inspect.getargspec(arguments.func)
        else:
            argspec = inspect.getfullargspec(arguments.func)
        vargs = []
        for arg in argspec.args:
            vargs.append(getattr(arguments, arg))
        if argspec.varargs:
            vargs.extend(getattr(arguments, argspec.varargs))
        output = arguments.func(*vargs)
        if getattr(arguments.func, '_cli_test_command', False):
            self.exit_code = 0 if output else 1
            output = ''
        if getattr(arguments.func, '_cli_no_output', False):
            output = ''
        self.formatter.format_output(output, arguments.format)
        if charmhelpers.core.unitdata._KV:
            charmhelpers.core.unitdata._KV.flush()


cmdline = CommandLine()


def describe_arguments(func):
    """
    Analyze a function's signature and return a data structure suitable for
    passing in as arguments to an argparse parser's add_argument() method."""

    if six.PY2:
        argspec = inspect.getargspec(func)
    else:
        argspec = inspect.getfullargspec(func)
    # we should probably raise an exception somewhere if func includes **kwargs
    if argspec.defaults:
        positional_args = argspec.args[:-len(argspec.defaults)]
        keyword_names = argspec.args[-len(argspec.defaults):]
        for arg, default in zip(keyword_names, argspec.defaults):
            yield ('--{}'.format(arg),), {'default': default}
    else:
        positional_args = argspec.args

    for arg in positional_args:
        yield (arg,), {}
    if argspec.varargs:
        yield (argspec.varargs,), {'nargs': '*'}



import contextlib
import copy
import io
import os
import mock
import six
import unittest

from charmhelpers.contrib.openstack import policyd


if not six.PY3:
    builtin_open = '__builtin__.open'
else:
    builtin_open = 'builtins.open'


class PolicydTests(unittest.TestCase):
    def setUp(self):
        super(PolicydTests, self).setUp()

    def test_is_policyd_override_valid_on_this_release(self):
        self.assertTrue(
            policyd.is_policyd_override_valid_on_this_release("queens"))
        self.assertTrue(
            policyd.is_policyd_override_valid_on_this_release("rocky"))
        self.assertFalse(
            policyd.is_policyd_override_valid_on_this_release("pike"))

    @mock.patch.object(policyd, "clean_policyd_dir_for")
    @mock.patch.object(policyd, "remove_policy_success_file")
    @mock.patch.object(policyd, "process_policy_resource_file")
    @mock.patch.object(policyd, "get_policy_resource_filename")
    @mock.patch.object(policyd, "is_policyd_override_valid_on_this_release")
    @mock.patch.object(policyd, "_policy_success_file")
    @mock.patch("os.path.isfile")
    @mock.patch.object(policyd.hookenv, "config")
    @mock.patch("charmhelpers.core.hookenv.log")
    def test_maybe_do_policyd_overrides(
        self,
        mock_log,
        mock_config,
        mock_isfile,
        mock__policy_success_file,
        mock_is_policyd_override_valid_on_this_release,
        mock_get_policy_resource_filename,
        mock_process_policy_resource_file,
        mock_remove_policy_success_file,
        mock_clean_policyd_dir_for,
    ):
        mock_isfile.return_value = False
        mock__policy_success_file.return_value = "s-return"
        # test success condition
        mock_config.return_value = {policyd.POLICYD_CONFIG_NAME: True}
        mock_is_policyd_override_valid_on_this_release.return_value = True
        mock_get_policy_resource_filename.return_value = "resource.zip"
        mock_process_policy_resource_file.return_value = True
        mod_fn = mock.Mock()
        restart_handler = mock.Mock()
        policyd.maybe_do_policyd_overrides(
            "arelease", "aservice", ["a"], ["b"], mod_fn, restart_handler)
        mock_is_policyd_override_valid_on_this_release.assert_called_once_with(
            "arelease")
        mock_get_policy_resource_filename.assert_called_once_with()
        mock_process_policy_resource_file.assert_called_once_with(
            "resource.zip", "aservice", ["a"], ["b"], mod_fn)
        restart_handler.assert_called_once_with()
        # test process_policy_resource_file is not called if not valid on the
        # release.
        mock_process_policy_resource_file.reset_mock()
        restart_handler.reset_mock()
        mock_is_policyd_override_valid_on_this_release.return_value = False
        policyd.maybe_do_policyd_overrides(
            "arelease", "aservice", ["a"], ["b"], mod_fn, restart_handler)
        mock_process_policy_resource_file.assert_not_called()
        restart_handler.assert_not_called()
        # test restart_handler is not called if not needed.
        mock_is_policyd_override_valid_on_this_release.return_value = True
        mock_process_policy_resource_file.return_value = False
        policyd.maybe_do_policyd_overrides(
            "arelease", "aservice", ["a"], ["b"], mod_fn, restart_handler)
        mock_process_policy_resource_file.assert_called_once_with(
            "resource.zip", "aservice", ["a"], ["b"], mod_fn)
        restart_handler.assert_not_called()
        # test that directory gets cleaned if the config is not set
        mock_config.return_value = {policyd.POLICYD_CONFIG_NAME: False}
        mock_process_policy_resource_file.reset_mock()
        policyd.maybe_do_policyd_overrides(
            "arelease", "aservice", ["a"], ["b"], mod_fn, restart_handler)
        mock_process_policy_resource_file.assert_not_called()
        mock_remove_policy_success_file.assert_called_once_with()
        mock_clean_policyd_dir_for.assert_called_once_with(
            "aservice", ["a"], user='aservice', group='aservice')

    @mock.patch.object(policyd, "maybe_do_policyd_overrides")
    def test_maybe_do_policyd_overrides_with_config_changed(
        self,
        mock_maybe_do_policyd_overrides,
    ):
        mod_fn = mock.Mock()
        restart_handler = mock.Mock()
        policyd.maybe_do_policyd_overrides_on_config_changed(
            "arelease", "aservice", ["a"], ["b"], mod_fn, restart_handler)
        mock_maybe_do_policyd_overrides.assert_called_once_with(
            "arelease", "aservice", ["a"], ["b"], mod_fn, restart_handler,
            config_changed=True)

    @mock.patch("charmhelpers.core.hookenv.resource_get")
    def test_get_policy_resource_filename(self, mock_resource_get):
        mock_resource_get.return_value = "test-file"
        self.assertEqual(policyd.get_policy_resource_filename(),
                         "test-file")
        mock_resource_get.assert_called_once_with(
            policyd.POLICYD_RESOURCE_NAME)

        # check that if an error is raised, that None is returned.
        def go_bang(*args):
            raise Exception("bang")

        mock_resource_get.side_effect = go_bang
        self.assertEqual(policyd.get_policy_resource_filename(), None)

    @mock.patch.object(policyd, "_yamlfiles")
    @mock.patch.object(policyd.zipfile, "ZipFile")
    def test_open_and_filter_yaml_files(self, mock_ZipFile, mock__yamlfiles):
        mock__yamlfiles.return_value = [
            ("file1", ".yaml", "file1.yaml", None),
            ("file2", ".yml", "file2.YML", None)]
        mock_ZipFile.return_value.__enter__.return_value = "zfp"
        # test a valid zip file
        with policyd.open_and_filter_yaml_files("some-file") as (zfp, files):
            self.assertEqual(zfp, "zfp")
            mock_ZipFile.assert_called_once_with("some-file", "r")
            self.assertEqual(files, [
                ("file1", ".yaml", "file1.yaml", None),
                ("file2", ".yml", "file2.YML", None)])
        # ensure that there must be at least one file.
        mock__yamlfiles.return_value = []
        with self.assertRaises(policyd.BadPolicyZipFile):
            with policyd.open_and_filter_yaml_files("some-file"):
                pass
        # ensure that it picks up duplicates
        mock__yamlfiles.return_value = [
            ("file1", ".yaml", "file1.yaml", None),
            ("file2", ".yml", "file2.yml", None),
            ("file1", ".yml", "file1.yml", None)]
        with self.assertRaises(policyd.BadPolicyZipFile):
            with policyd.open_and_filter_yaml_files("some-file"):
                pass

    def test__yamlfiles(self):
        class MockZipFile(object):
            def __init__(self, infolist):
                self._infolist = infolist

            def infolist(self):
                return self._infolist

        class MockInfoListItem(object):
            def __init__(self, is_dir, filename):
                self.filename = filename
                self._is_dir = is_dir

            def is_dir(self):
                return self._is_dir

            def __repr__(self):
                return "MockInfoListItem({}, {})".format(self._is_dir,
                                                         self.filename)

        zipfile = MockZipFile([
            MockInfoListItem(False, "file1.yaml"),
            MockInfoListItem(False, "file2.md"),
            MockInfoListItem(False, "file3.YML"),
            MockInfoListItem(False, "file4.Yaml"),
            MockInfoListItem(True, "file5"),
            MockInfoListItem(True, "file6.yaml"),
            MockInfoListItem(False, "file7"),
            MockInfoListItem(False, "file8.j2")])

        self.assertEqual(list(policyd._yamlfiles(zipfile)),
                         [("file1", ".yaml", "file1.yaml", mock.ANY),
                          ("file3", ".yml", "file3.YML", mock.ANY),
                          ("file4", ".yaml", "file4.Yaml", mock.ANY),
                          ("file8", ".j2", "file8.j2", mock.ANY)])

    @mock.patch.object(policyd.yaml, "safe_load")
    def test_read_and_validate_yaml(self, mock_safe_load):
        # test a valid document
        good_doc = {
            "key1": "rule1",
            "key2": "rule2",
        }
        mock_safe_load.return_value = copy.deepcopy(good_doc)
        doc = policyd.read_and_validate_yaml("test-stream")
        self.assertEqual(doc, good_doc)
        mock_safe_load.assert_called_once_with("test-stream")
        # test an invalid document - return a string
        mock_safe_load.return_value = "wrong"
        with self.assertRaises(policyd.BadPolicyYamlFile):
            policyd.read_and_validate_yaml("test-stream")
        # test for black-listed keys
        with self.assertRaises(policyd.BadPolicyYamlFile):
            mock_safe_load.return_value = copy.deepcopy(good_doc)
            policyd.read_and_validate_yaml("test-stream", ["key1"])
        # test for non string keys
        bad_key_doc = {
            (1,): "rule1",
            "key2": "rule2",
        }
        with self.assertRaises(policyd.BadPolicyYamlFile):
            mock_safe_load.return_value = copy.deepcopy(bad_key_doc)
            policyd.read_and_validate_yaml("test-stream", ["key1"])
        # test for non string values (i.e. no nested keys)
        bad_key_doc2 = {
            "key1": "rule1",
            "key2": {"sub_key": "rule2"},
        }
        with self.assertRaises(policyd.BadPolicyYamlFile):
            mock_safe_load.return_value = copy.deepcopy(bad_key_doc2)
            policyd.read_and_validate_yaml("test-stream", ["key1"])

    def test_policyd_dir_for(self):
        self.assertEqual(policyd.policyd_dir_for('thing'),
                         "/etc/thing/policy.d")

    @mock.patch.object(policyd.hookenv, 'log')
    @mock.patch("os.remove")
    @mock.patch("shutil.rmtree")
    @mock.patch("charmhelpers.core.host.mkdir")
    @mock.patch("os.path.exists")
    @mock.patch.object(policyd, "policyd_dir_for")
    def test_clean_policyd_dir_for(self,
                                   mock_policyd_dir_for,
                                   mock_os_path_exists,
                                   mock_mkdir,
                                   mock_shutil_rmtree,
                                   mock_os_remove,
                                   mock_log):
        if hasattr(os, 'scandir'):
            mock_scan_dir_parts = (mock.patch, ["os.scandir"])
        else:
            mock_scan_dir_parts = (mock.patch.object,
                                   [policyd, "_fallback_scandir"])

        class MockDirEntry(object):
            def __init__(self, path, is_dir):
                self.path = path
                self._is_dir = is_dir

            def is_dir(self):
                return self._is_dir

        # list of scanned objects
        directory_contents = [
            MockDirEntry("one", False),
            MockDirEntry("two", False),
            MockDirEntry("three", True),
            MockDirEntry("four", False)]

        mock_policyd_dir_for.return_value = "the-path"

        # Initial conditions
        mock_os_path_exists.return_value = False

        # call the function
        with mock_scan_dir_parts[0](*mock_scan_dir_parts[1]) as \
                mock_os_scandir:
            mock_os_scandir.return_value = directory_contents
            policyd.clean_policyd_dir_for("aservice")

        # check it did the right thing
        mock_policyd_dir_for.assert_called_once_with("aservice")
        mock_os_path_exists.assert_called_once_with("the-path")
        mock_mkdir.assert_called_once_with("the-path",
                                           owner="aservice",
                                           group="aservice",
                                           perms=0o775)
        mock_shutil_rmtree.assert_called_once_with("three")
        mock_os_remove.assert_has_calls([
            mock.call("one"), mock.call("two"), mock.call("four")])

        # check also that we can omit paths ... reset everything
        mock_os_remove.reset_mock()
        mock_shutil_rmtree.reset_mock()
        mock_os_path_exists.reset_mock()
        mock_os_path_exists.return_value = True
        mock_mkdir.reset_mock()

        with mock_scan_dir_parts[0](*mock_scan_dir_parts[1]) as \
                mock_os_scandir:
            mock_os_scandir.return_value = directory_contents
            policyd.clean_policyd_dir_for("aservice",
                                          keep_paths=["one", "three"])

        # verify all worked as we expected
        mock_mkdir.assert_not_called()
        mock_shutil_rmtree.assert_not_called()
        mock_os_remove.assert_has_calls([mock.call("two"), mock.call("four")])

    def test_path_for_policy_file(self):
        self.assertEqual(policyd.path_for_policy_file('this', 'that'),
                         "/etc/this/policy.d/that.yaml")

    @mock.patch("charmhelpers.core.hookenv.charm_dir")
    def test__policy_success_file(self, mock_charm_dir):
        mock_charm_dir.return_value = "/this"
        self.assertEqual(policyd._policy_success_file(),
                         "/this/{}".format(policyd.POLICYD_SUCCESS_FILENAME))

    @mock.patch("os.remove")
    @mock.patch.object(policyd, "_policy_success_file")
    def test_remove_policy_success_file(self, mock_file, mock_os_remove):
        mock_file.return_value = "the-path"
        policyd.remove_policy_success_file()
        mock_os_remove.assert_called_once_with("the-path")

        # now test that failure doesn't fail the function
        def go_bang(*args):
            raise Exception("bang")

        mock_os_remove.side_effect = go_bang
        policyd.remove_policy_success_file()

    @mock.patch("os.path.isfile")
    @mock.patch.object(policyd, "_policy_success_file")
    def test_policyd_status_message_prefix(self, mock_file, mock_is_file):
        mock_file.return_value = "the-path"
        mock_is_file.return_value = True
        self.assertEqual(policyd.policyd_status_message_prefix(), "PO:")
        mock_is_file.return_value = False
        self.assertEqual(
            policyd.policyd_status_message_prefix(), "PO (broken):")

    @mock.patch("yaml.dump")
    @mock.patch.object(policyd, "_policy_success_file")
    @mock.patch.object(policyd.hookenv, "log")
    @mock.patch.object(policyd, "read_and_validate_yaml")
    @mock.patch.object(policyd, "path_for_policy_file")
    @mock.patch.object(policyd, "clean_policyd_dir_for")
    @mock.patch.object(policyd, "remove_policy_success_file")
    @mock.patch.object(policyd, "open_and_filter_yaml_files")
    @mock.patch.object(policyd.ch_host, 'write_file')
    @mock.patch.object(policyd, "maybe_create_directory_for")
    def test_process_policy_resource_file(
        self,
        mock_maybe_create_directory_for,
        mock_write_file,
        mock_open_and_filter_yaml_files,
        mock_remove_policy_success_file,
        mock_clean_policyd_dir_for,
        mock_path_for_policy_file,
        mock_read_and_validate_yaml,
        mock_log,
        mock__policy_success_file,
        mock_yaml_dump,
    ):
        mock_zfp = mock.MagicMock()
        mod_fn = mock.Mock()
        mock_path_for_policy_file.side_effect = lambda s, n: s + "/" + n
        gen = [
            ("file1", ".yaml", "file1.yaml", "file1-zipinfo"),
            ("file2", ".yml", "file2.yml", "file2-zipinfo")]
        mock_open_and_filter_yaml_files.return_value.__enter__.return_value = \
            (mock_zfp, gen)
        # first verify that we can blacklist a file
        res = policyd.process_policy_resource_file(
            "resource.zip", "aservice", ["aservice/file1"], [], mod_fn)
        self.assertFalse(res)
        mock_remove_policy_success_file.assert_called_once_with()
        mock_clean_policyd_dir_for.assert_has_calls([
            mock.call("aservice",
                      ["aservice/file1"],
                      user='aservice',
                      group='aservice'),
            mock.call("aservice",
                      ["aservice/file1"],
                      user='aservice',
                      group='aservice')])
        mock_zfp.open.assert_not_called()
        mod_fn.assert_not_called()
        mock_log.assert_any_call("Processing resource.zip failed: policy.d"
                                 " name aservice/file1 is blacklisted",
                                 level=policyd.POLICYD_LOG_LEVEL_DEFAULT)

        # now test for success
        @contextlib.contextmanager
        def _patch_open():
            '''Patch open() to allow mocking both open() itself and the file that is
            yielded.

            Yields the mock for "open" and "file", respectively.'''
            mock_open = mock.MagicMock(spec=open)
            mock_file = mock.MagicMock(spec=io.FileIO)

            with mock.patch(builtin_open, mock_open):
                yield mock_open, mock_file

        mock_clean_policyd_dir_for.reset_mock()
        mock_zfp.reset_mock()
        mock_fp = mock.MagicMock()
        mock_fp.read.return_value = '{"rule1": "value1"}'
        mock_zfp.open.return_value.__enter__.return_value = mock_fp
        gen = [("file1", ".j2", "file1.j2", "file1-zipinfo")]
        mock_open_and_filter_yaml_files.return_value.__enter__.return_value = \
            (mock_zfp, gen)
        mock_read_and_validate_yaml.return_value = {"rule1": "modded_value1"}
        mod_fn.return_value = '{"rule1": "modded_value1"}'
        mock__policy_success_file.return_value = "policy-success-file"
        mock_yaml_dump.return_value = "dumped-file"
        with _patch_open() as (mock_open, mock_file):
            res = policyd.process_policy_resource_file(
                "resource.zip", "aservice", [], ["key"], mod_fn)
            self.assertTrue(res)
            # mock_open.assert_any_call("aservice/file1", "wt")
            mock_write_file.assert_called_once_with(
                "aservice/file1",
                b'dumped-file',
                "aservice",
                "aservice")
            mock_open.assert_any_call("policy-success-file", "w")
            mock_yaml_dump.assert_called_once_with({"rule1": "modded_value1"})
        mock_zfp.open.assert_called_once_with("file1-zipinfo")
        mock_read_and_validate_yaml.assert_called_once_with(
            '{"rule1": "modded_value1"}', ["key"])
        mod_fn.assert_called_once_with('{"rule1": "value1"}')

        # raise a BadPolicyZipFile if we have a template, but there is no
        # template function
        mock_log.reset_mock()
        with _patch_open() as (mock_open, mock_file):
            res = policyd.process_policy_resource_file(
                "resource.zip", "aservice", [], ["key"],
                template_function=None)
            self.assertFalse(res)
        mock_log.assert_any_call(
            "Processing resource.zip failed: Template file1.j2 "
            "but no template_function is available",
            level=policyd.POLICYD_LOG_LEVEL_DEFAULT)

        # raise the IOError to validate that code path
        def raise_ioerror(*args):
            raise IOError("bang")

        mock_open_and_filter_yaml_files.side_effect = raise_ioerror
        mock_log.reset_mock()
        res = policyd.process_policy_resource_file(
            "resource.zip", "aservice", [], ["key"], mod_fn)
        self.assertFalse(res, False)
        mock_log.assert_any_call(
            "File resource.zip failed with IOError.  "
            "This really shouldn't happen -- error: bang",
            level=policyd.POLICYD_LOG_LEVEL_DEFAULT)
        # raise a general exception, so that is caught and logged too.

        def raise_exception(*args):
            raise Exception("bang2")

        mock_open_and_filter_yaml_files.reset_mock()
        mock_open_and_filter_yaml_files.side_effect = raise_exception
        mock_log.reset_mock()
        res = policyd.process_policy_resource_file(
            "resource.zip", "aservice", [], ["key"], mod_fn)
        self.assertFalse(res, False)
        mock_log.assert_any_call(
            "General Exception(bang2) during policyd processing",
            level=policyd.POLICYD_LOG_LEVEL_DEFAULT)



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from collections import OrderedDict

from charmhelpers.core.hookenv import (
    config,
    log,
    DEBUG,
    WARNING,
)
from charmhelpers.contrib.hardening.host.checks import run_os_checks
from charmhelpers.contrib.hardening.ssh.checks import run_ssh_checks
from charmhelpers.contrib.hardening.mysql.checks import run_mysql_checks
from charmhelpers.contrib.hardening.apache.checks import run_apache_checks

_DISABLE_HARDENING_FOR_UNIT_TEST = False


def harden(overrides=None):
    """Hardening decorator.

    This is the main entry point for running the hardening stack. In order to
    run modules of the stack you must add this decorator to charm hook(s) and
    ensure that your charm config.yaml contains the 'harden' option set to
    one or more of the supported modules. Setting these will cause the
    corresponding hardening code to be run when the hook fires.

    This decorator can and should be applied to more than one hook or function
    such that hardening modules are called multiple times. This is because
    subsequent calls will perform auditing checks that will report any changes
    to resources hardened by the first run (and possibly perform compliance
    actions as a result of any detected infractions).

    :param overrides: Optional list of stack modules used to override those
                      provided with 'harden' config.
    :returns: Returns value returned by decorated function once executed.
    """
    if overrides is None:
        overrides = []

    def _harden_inner1(f):
        # As this has to be py2.7 compat, we can't use nonlocal.  Use a trick
        # to capture the dictionary that can then be updated.
        _logged = {'done': False}

        def _harden_inner2(*args, **kwargs):
            # knock out hardening via a config var; normally it won't get
            # disabled.
            if _DISABLE_HARDENING_FOR_UNIT_TEST:
                return f(*args, **kwargs)
            if not _logged['done']:
                log("Hardening function '%s'" % (f.__name__), level=DEBUG)
                _logged['done'] = True
            RUN_CATALOG = OrderedDict([('os', run_os_checks),
                                       ('ssh', run_ssh_checks),
                                       ('mysql', run_mysql_checks),
                                       ('apache', run_apache_checks)])

            enabled = overrides[:] or (config("harden") or "").split()
            if enabled:
                modules_to_run = []
                # modules will always be performed in the following order
                for module, func in six.iteritems(RUN_CATALOG):
                    if module in enabled:
                        enabled.remove(module)
                        modules_to_run.append(func)

                if enabled:
                    log("Unknown hardening modules '%s' - ignoring" %
                        (', '.join(enabled)), level=WARNING)

                for hardener in modules_to_run:
                    log("Executing hardening module '%s'" %
                        (hardener.__name__), level=DEBUG)
                    hardener()
            else:
                log("No hardening applied to '%s'" % (f.__name__), level=DEBUG)

            return f(*args, **kwargs)
        return _harden_inner2

    return _harden_inner1



#!/usr/bin/env python

# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import subprocess
from charmhelpers.core import hookenv


def default_execd_dir():
    return os.path.join(os.environ['CHARM_DIR'], 'exec.d')


def execd_module_paths(execd_dir=None):
    """Generate a list of full paths to modules within execd_dir."""
    if not execd_dir:
        execd_dir = default_execd_dir()

    if not os.path.exists(execd_dir):
        return

    for subpath in os.listdir(execd_dir):
        module = os.path.join(execd_dir, subpath)
        if os.path.isdir(module):
            yield module


def execd_submodule_paths(command, execd_dir=None):
    """Generate a list of full paths to the specified command within exec_dir.
    """
    for module_path in execd_module_paths(execd_dir):
        path = os.path.join(module_path, command)
        if os.access(path, os.X_OK) and os.path.isfile(path):
            yield path


def execd_run(command, execd_dir=None, die_on_error=True, stderr=subprocess.STDOUT):
    """Run command for each module within execd_dir which defines it."""
    for submodule_path in execd_submodule_paths(command, execd_dir):
        try:
            subprocess.check_output(submodule_path, stderr=stderr,
                                    universal_newlines=True)
        except subprocess.CalledProcessError as e:
            hookenv.log("Error ({}) running  {}. Output: {}".format(
                e.returncode, e.cmd, e.output))
            if die_on_error:
                sys.exit(e.returncode)


def execd_preinstall(execd_dir=None):
    """Run charm-pre-install for each module within execd_dir."""
    execd_run('charm-pre-install', execd_dir=execd_dir)



# Copyright 2016 Canonical Ltd
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit test helpers from https://github.com/openstack/charms.openstack/"""

import contextlib
import io
import mock
import unittest


@contextlib.contextmanager
def patch_open():
    '''Patch open() to allow mocking both open() itself and the file that is
    yielded.

    Yields the mock for "open" and "file", respectively.'''
    mock_open = mock.MagicMock(spec=open)
    mock_file = mock.MagicMock(spec=io.FileIO)

    @contextlib.contextmanager
    def stub_open(*args, **kwargs):
        mock_open(*args, **kwargs)
        yield mock_file

    with mock.patch('builtins.open', stub_open):
        yield mock_open, mock_file


class BaseTestCase(unittest.TestCase):

    def setUp(self):
        self._patches = {}
        self._patches_start = {}

    def tearDown(self):
        for k, v in self._patches.items():
            v.stop()
            setattr(self, k, None)
        self._patches = None
        self._patches_start = None

    def patch_object(self, obj, attr, return_value=None, name=None, new=None,
                     **kwargs):
        if name is None:
            name = attr
        if new is not None:
            mocked = mock.patch.object(obj, attr, new=new, **kwargs)
        else:
            mocked = mock.patch.object(obj, attr, **kwargs)
        self._patches[name] = mocked
        started = mocked.start()
        if new is None:
            started.return_value = return_value
        self._patches_start[name] = started
        setattr(self, name, started)

    def patch(self, item, return_value=None, name=None, new=None, **kwargs):
        if name is None:
            raise RuntimeError("Must pass 'name' to .patch()")
        if new is not None:
            mocked = mock.patch(item, new=new, **kwargs)
        else:
            mocked = mock.patch(item, **kwargs)
        self._patches[name] = mocked
        started = mocked.start()
        if new is None:
            started.return_value = return_value
        self._patches_start[name] = started
        setattr(self, name, started)



from mock import patch
import unittest
import json

from charmhelpers.contrib.openstack.ha import utils as ha

IFACE_LOOKUPS = {
    '10.5.100.1': 'eth1',
    'ffff::1': 'eth1',
    'ffaa::1': 'eth2',
}

NETMASK_LOOKUPS = {
    '10.5.100.1': '255.255.255.0',
    'ffff::1': '64',
    'ffaa::1': '32',
}


class HATests(unittest.TestCase):
    def setUp(self):
        super(HATests, self).setUp()
        [self._patch(m) for m in [
            'charm_name',
            'config',
            'relation_set',
            'resolve_address',
            'status_set',
            'get_hacluster_config',
            'get_iface_for_address',
            'get_netmask_for_address',
        ]]
        self.resources = {'res_test_haproxy': 'lsb:haproxy'}
        self.resource_params = {'res_test_haproxy': 'op monitor interval="5s"'}
        self.conf = {}
        self.config.side_effect = lambda key: self.conf.get(key)
        self.maxDiff = None
        self.get_iface_for_address.side_effect = \
            lambda x: IFACE_LOOKUPS.get(x)
        self.get_netmask_for_address.side_effect = \
            lambda x: NETMASK_LOOKUPS.get(x)

    def _patch(self, method):
        _m = patch.object(ha, method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        setattr(self, method, mock)

    @patch.object(ha, 'log', lambda *args, **kwargs: None)
    @patch.object(ha, 'assert_charm_supports_dns_ha')
    def test_update_dns_ha_resource_params_none(self,
                                                assert_charm_supports_dns_ha):
        self.conf = {
            'os-admin-hostname': None,
            'os-internal-hostname': None,
            'os-public-hostname': None,
        }

        with self.assertRaises(ha.DNSHAException):
            ha.update_dns_ha_resource_params(
                relation_id='ha:1',
                resources=self.resources,
                resource_params=self.resource_params)

    @patch.object(ha, 'log', lambda *args, **kwargs: None)
    @patch.object(ha, 'assert_charm_supports_dns_ha')
    def test_update_dns_ha_resource_params_one(self,
                                               assert_charm_supports_dns_ha):
        EXPECTED_RESOURCES = {'res_test_public_hostname': 'ocf:maas:dns',
                              'res_test_haproxy': 'lsb:haproxy'}
        EXPECTED_RESOURCE_PARAMS = {
            'res_test_public_hostname': ('params fqdn="test.maas" '
                                         'ip_address="10.0.0.1"'),
            'res_test_haproxy': 'op monitor interval="5s"'}

        self.conf = {
            'os-admin-hostname': None,
            'os-internal-hostname': None,
            'os-public-hostname': 'test.maas',
        }

        self.charm_name.return_value = 'test'
        self.resolve_address.return_value = '10.0.0.1'
        ha.update_dns_ha_resource_params(relation_id='ha:1',
                                         resources=self.resources,
                                         resource_params=self.resource_params)
        self.assertEqual(self.resources, EXPECTED_RESOURCES)
        self.assertEqual(self.resource_params, EXPECTED_RESOURCE_PARAMS)
        self.relation_set.assert_called_with(
            groups={'grp_test_hostnames': 'res_test_public_hostname'},
            relation_id='ha:1')

    @patch.object(ha, 'log', lambda *args, **kwargs: None)
    @patch.object(ha, 'assert_charm_supports_dns_ha')
    def test_update_dns_ha_resource_params_all(self,
                                               assert_charm_supports_dns_ha):
        EXPECTED_RESOURCES = {'res_test_admin_hostname': 'ocf:maas:dns',
                              'res_test_int_hostname': 'ocf:maas:dns',
                              'res_test_public_hostname': 'ocf:maas:dns',
                              'res_test_haproxy': 'lsb:haproxy'}
        EXPECTED_RESOURCE_PARAMS = {
            'res_test_admin_hostname': ('params fqdn="test.admin.maas" '
                                        'ip_address="10.0.0.1"'),
            'res_test_int_hostname': ('params fqdn="test.internal.maas" '
                                      'ip_address="10.0.0.1"'),
            'res_test_public_hostname': ('params fqdn="test.public.maas" '
                                         'ip_address="10.0.0.1"'),
            'res_test_haproxy': 'op monitor interval="5s"'}

        self.conf = {
            'os-admin-hostname': 'test.admin.maas',
            'os-internal-hostname': 'test.internal.maas',
            'os-public-hostname': 'test.public.maas',
        }

        self.charm_name.return_value = 'test'
        self.resolve_address.return_value = '10.0.0.1'
        ha.update_dns_ha_resource_params(relation_id='ha:1',
                                         resources=self.resources,
                                         resource_params=self.resource_params)
        self.assertEqual(self.resources, EXPECTED_RESOURCES)
        self.assertEqual(self.resource_params, EXPECTED_RESOURCE_PARAMS)
        self.relation_set.assert_called_with(
            groups={'grp_test_hostnames':
                    ('res_test_admin_hostname '
                     'res_test_int_hostname '
                     'res_test_public_hostname')},
            relation_id='ha:1')

    @patch.object(ha, 'lsb_release')
    def test_assert_charm_supports_dns_ha(self, lsb_release):
        lsb_release.return_value = {'DISTRIB_RELEASE': '16.04'}
        self.assertTrue(ha.assert_charm_supports_dns_ha())

    @patch.object(ha, 'lsb_release')
    def test_assert_charm_supports_dns_ha_exception(self, lsb_release):
        lsb_release.return_value = {'DISTRIB_RELEASE': '12.04'}
        self.assertRaises(ha.DNSHAException,
                          lambda: ha.assert_charm_supports_dns_ha())

    @patch.object(ha, 'expected_related_units')
    def tests_expect_ha(self, expected_related_units):
        expected_related_units.return_value = (x for x in [])
        self.conf = {'vip': None,
                     'dns-ha': None}
        self.assertFalse(ha.expect_ha())

        expected_related_units.return_value = (x for x in ['hacluster-unit/0',
                                                           'hacluster-unit/1',
                                                           'hacluster-unit/2'])
        self.conf = {'vip': None,
                     'dns-ha': None}
        self.assertTrue(ha.expect_ha())

        expected_related_units.side_effect = NotImplementedError
        self.conf = {'vip': '10.0.0.1',
                     'dns-ha': None}
        self.assertTrue(ha.expect_ha())

        self.conf = {'vip': None,
                     'dns-ha': True}
        self.assertTrue(ha.expect_ha())

    def test_get_vip_settings(self):
        self.assertEqual(
            ha.get_vip_settings('10.5.100.1'),
            ('eth1', '255.255.255.0', False))

    def test_get_vip_settings_fallback(self):
        self.conf = {'vip_iface': 'eth3',
                     'vip_cidr': '255.255.0.0'}
        self.assertEqual(
            ha.get_vip_settings('192.168.100.1'),
            ('eth3', '255.255.0.0', True))

    def test_update_hacluster_vip_single_vip(self):
        self.get_hacluster_config.return_value = {
            'vip': '10.5.100.1'
        }
        test_data = {'resources': {}, 'resource_params': {}}
        expected = {
            'delete_resources': ['res_testservice_eth1_vip'],
            'groups': {
                'grp_testservice_vips': 'res_testservice_242d562_vip'
            },
            'resource_params': {
                'res_testservice_242d562_vip':
                    ('params ip="10.5.100.1" op monitor '
                     'timeout="20s" interval="10s" depth="0"')
            },
            'resources': {
                'res_testservice_242d562_vip': 'ocf:heartbeat:IPaddr2'
            }
        }
        ha.update_hacluster_vip('testservice', test_data)
        self.assertEqual(test_data, expected)

    def test_update_hacluster_vip_single_vip_fallback(self):
        self.get_hacluster_config.return_value = {
            'vip': '10.5.100.1'
        }
        test_data = {'resources': {}, 'resource_params': {}}
        expected = {
            'delete_resources': ['res_testservice_eth1_vip'],
            'groups': {
                'grp_testservice_vips': 'res_testservice_242d562_vip'
            },
            'resource_params': {
                'res_testservice_242d562_vip':
                    ('params ip="10.5.100.1" op monitor '
                     'timeout="20s" interval="10s" depth="0"')
            },
            'resources': {
                'res_testservice_242d562_vip': 'ocf:heartbeat:IPaddr2'
            }
        }
        ha.update_hacluster_vip('testservice', test_data)
        self.assertEqual(test_data, expected)

    def test_update_hacluster_config_vip(self):
        self.get_iface_for_address.side_effect = lambda x: None
        self.get_netmask_for_address.side_effect = lambda x: None
        self.conf = {'vip_iface': 'eth1',
                     'vip_cidr': '255.255.255.0'}
        self.get_hacluster_config.return_value = {
            'vip': '10.5.100.1'
        }
        test_data = {'resources': {}, 'resource_params': {}}
        expected = {
            'delete_resources': ['res_testservice_eth1_vip'],
            'groups': {
                'grp_testservice_vips': 'res_testservice_242d562_vip'
            },
            'resource_params': {
                'res_testservice_242d562_vip': (
                    'params ip="10.5.100.1" cidr_netmask="255.255.255.0" '
                    'nic="eth1" op monitor timeout="20s" '
                    'interval="10s" depth="0"')

            },
            'resources': {
                'res_testservice_242d562_vip': 'ocf:heartbeat:IPaddr2'
            }
        }
        ha.update_hacluster_vip('testservice', test_data)
        self.assertEqual(test_data, expected)

    def test_update_hacluster_vip_multiple_vip(self):
        self.get_hacluster_config.return_value = {
            'vip': '10.5.100.1 ffff::1 ffaa::1'
        }
        test_data = {'resources': {}, 'resource_params': {}}
        expected = {
            'groups': {
                'grp_testservice_vips': ('res_testservice_242d562_vip '
                                         'res_testservice_856d56f_vip '
                                         'res_testservice_f563c5d_vip')
            },
            'delete_resources': ['res_testservice_eth1_vip',
                                 'res_testservice_eth1_vip_ipv6addr',
                                 'res_testservice_eth2_vip'],
            'resource_params': {
                'res_testservice_242d562_vip':
                    ('params ip="10.5.100.1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
                'res_testservice_856d56f_vip':
                    ('params ipv6addr="ffff::1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
                'res_testservice_f563c5d_vip':
                    ('params ipv6addr="ffaa::1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
            },
            'resources': {
                'res_testservice_242d562_vip': 'ocf:heartbeat:IPaddr2',
                'res_testservice_856d56f_vip': 'ocf:heartbeat:IPv6addr',
                'res_testservice_f563c5d_vip': 'ocf:heartbeat:IPv6addr',
            }
        }
        ha.update_hacluster_vip('testservice', test_data)
        self.assertEqual(test_data, expected)

    def test_generate_ha_relation_data_haproxy_disabled(self):
        self.get_hacluster_config.return_value = {
            'vip': '10.5.100.1 ffff::1 ffaa::1'
        }
        extra_settings = {
            'colocations': {'vip_cauth': 'inf: res_nova_cauth grp_nova_vips'},
            'init_services': {'res_nova_cauth': 'nova-cauth'},
            'delete_resources': ['res_ceilometer_polling'],
            'groups': {'grp_testservice_wombles': 'res_testservice_orinoco'},
        }
        expected = {
            'colocations': {'vip_cauth': 'inf: res_nova_cauth grp_nova_vips'},
            'groups': {
                'grp_testservice_vips': ('res_testservice_242d562_vip '
                                         'res_testservice_856d56f_vip '
                                         'res_testservice_f563c5d_vip'),
                'grp_testservice_wombles': 'res_testservice_orinoco'
            },
            'resource_params': {
                'res_testservice_242d562_vip':
                    ('params ip="10.5.100.1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
                'res_testservice_856d56f_vip':
                    ('params ipv6addr="ffff::1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
                'res_testservice_f563c5d_vip':
                    ('params ipv6addr="ffaa::1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
            },
            'resources': {
                'res_testservice_242d562_vip': 'ocf:heartbeat:IPaddr2',
                'res_testservice_856d56f_vip': 'ocf:heartbeat:IPv6addr',
                'res_testservice_f563c5d_vip': 'ocf:heartbeat:IPv6addr',
            },
            'clones': {},
            'init_services': {
                'res_nova_cauth': 'nova-cauth'
            },
            'delete_resources': ["res_ceilometer_polling",
                                 "res_testservice_eth1_vip",
                                 "res_testservice_eth1_vip_ipv6addr",
                                 "res_testservice_eth2_vip"],
        }
        expected = {
            'json_{}'.format(k): json.dumps(v, **ha.JSON_ENCODE_OPTIONS)
            for k, v in expected.items() if v
        }
        self.assertEqual(
            ha.generate_ha_relation_data('testservice',
                                         haproxy_enabled=False,
                                         extra_settings=extra_settings),
            expected)

    def test_generate_ha_relation_data(self):
        self.get_hacluster_config.return_value = {
            'vip': '10.5.100.1 ffff::1 ffaa::1'
        }
        extra_settings = {
            'colocations': {'vip_cauth': 'inf: res_nova_cauth grp_nova_vips'},
            'init_services': {'res_nova_cauth': 'nova-cauth'},
            'delete_resources': ['res_ceilometer_polling'],
            'groups': {'grp_testservice_wombles': 'res_testservice_orinoco'},
        }
        expected = {
            'colocations': {'vip_cauth': 'inf: res_nova_cauth grp_nova_vips'},
            'groups': {
                'grp_testservice_vips': ('res_testservice_242d562_vip '
                                         'res_testservice_856d56f_vip '
                                         'res_testservice_f563c5d_vip'),
                'grp_testservice_wombles': 'res_testservice_orinoco'
            },
            'resource_params': {
                'res_testservice_242d562_vip':
                    ('params ip="10.5.100.1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
                'res_testservice_856d56f_vip':
                    ('params ipv6addr="ffff::1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
                'res_testservice_f563c5d_vip':
                    ('params ipv6addr="ffaa::1" op monitor '
                     'timeout="20s" interval="10s" depth="0"'),
                'res_testservice_haproxy':
                    ('meta migration-threshold="INFINITY" failure-timeout="5s" '
                     'op monitor interval="5s"'),
            },
            'resources': {
                'res_testservice_242d562_vip': 'ocf:heartbeat:IPaddr2',
                'res_testservice_856d56f_vip': 'ocf:heartbeat:IPv6addr',
                'res_testservice_f563c5d_vip': 'ocf:heartbeat:IPv6addr',
                'res_testservice_haproxy': 'lsb:haproxy',
            },
            'clones': {
                'cl_testservice_haproxy': 'res_testservice_haproxy',
            },
            'init_services': {
                'res_testservice_haproxy': 'haproxy',
                'res_nova_cauth': 'nova-cauth'
            },
            'delete_resources': ["res_ceilometer_polling",
                                 "res_testservice_eth1_vip",
                                 "res_testservice_eth1_vip_ipv6addr",
                                 "res_testservice_eth2_vip"],
        }
        expected = {
            'json_{}'.format(k): json.dumps(v, **ha.JSON_ENCODE_OPTIONS)
            for k, v in expected.items() if v
        }
        self.assertEqual(
            ha.generate_ha_relation_data('testservice',
                                         extra_settings=extra_settings),
            expected)

    @patch.object(ha, 'log')
    @patch.object(ha, 'assert_charm_supports_dns_ha')
    def test_generate_ha_relation_data_dns_ha(self,
                                              assert_charm_supports_dns_ha,
                                              log):
        self.get_hacluster_config.return_value = {
            'vip': '10.5.100.1 ffff::1 ffaa::1'
        }
        self.conf = {
            'os-admin-hostname': 'test.admin.maas',
            'os-internal-hostname': 'test.internal.maas',
            'os-public-hostname': 'test.public.maas',
            'dns-ha': True,
        }
        self.resolve_address.return_value = '10.0.0.1'
        assert_charm_supports_dns_ha.return_value = True
        expected = {
            'groups': {
                'grp_testservice_hostnames': ('res_testservice_admin_hostname'
                                              ' res_testservice_int_hostname'
                                              ' res_testservice_public_hostname')
            },
            'resource_params': {
                'res_testservice_admin_hostname':
                    'params fqdn="test.admin.maas" ip_address="10.0.0.1"',
                'res_testservice_int_hostname':
                    'params fqdn="test.internal.maas" ip_address="10.0.0.1"',
                'res_testservice_public_hostname':
                    'params fqdn="test.public.maas" ip_address="10.0.0.1"',
                'res_testservice_haproxy':
                    ('meta migration-threshold="INFINITY" failure-timeout="5s" '
                     'op monitor interval="5s"'),
            },
            'resources': {
                'res_testservice_admin_hostname': 'ocf:maas:dns',
                'res_testservice_int_hostname': 'ocf:maas:dns',
                'res_testservice_public_hostname': 'ocf:maas:dns',
                'res_testservice_haproxy': 'lsb:haproxy',
            },
            'clones': {
                'cl_testservice_haproxy': 'res_testservice_haproxy',
            },
            'init_services': {
                'res_testservice_haproxy': 'haproxy'
            },
        }
        expected = {
            'json_{}'.format(k): json.dumps(v, **ha.JSON_ENCODE_OPTIONS)
            for k, v in expected.items() if v
        }
        self.assertEqual(ha.generate_ha_relation_data('testservice'),
                         expected)



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from charmhelpers.contrib.network.ip import (
    get_address_in_network,
    get_iface_addr,
    is_ip,
)
from charmhelpers.core.hookenv import (
    log,
    DEBUG,
)
from charmhelpers.fetch import (
    apt_install,
    apt_update,
)
from charmhelpers.core.host import (
    lsb_release,
    CompareHostReleases,
)
from charmhelpers.contrib.hardening.audits.file import (
    TemplatedFile,
    FileContentAudit,
)
from charmhelpers.contrib.hardening.ssh import TEMPLATES_DIR
from charmhelpers.contrib.hardening import utils


def get_audits():
    """Get SSH hardening config audits.

    :returns:  dictionary of audits
    """
    audits = [SSHConfig(), SSHDConfig(), SSHConfigFileContentAudit(),
              SSHDConfigFileContentAudit()]
    return audits


class SSHConfigContext(object):

    type = 'client'

    def get_macs(self, allow_weak_mac):
        if allow_weak_mac:
            weak_macs = 'weak'
        else:
            weak_macs = 'default'

        default = 'hmac-sha2-512,hmac-sha2-256,hmac-ripemd160'
        macs = {'default': default,
                'weak': default + ',hmac-sha1'}

        default = ('hmac-sha2-512-etm@openssh.com,'
                   'hmac-sha2-256-etm@openssh.com,'
                   'hmac-ripemd160-etm@openssh.com,umac-128-etm@openssh.com,'
                   'hmac-sha2-512,hmac-sha2-256,hmac-ripemd160')
        macs_66 = {'default': default,
                   'weak': default + ',hmac-sha1'}

        # Use newer ciphers on Ubuntu Trusty and above
        _release = lsb_release()['DISTRIB_CODENAME'].lower()
        if CompareHostReleases(_release) >= 'trusty':
            log("Detected Ubuntu 14.04 or newer, using new macs", level=DEBUG)
            macs = macs_66

        return macs[weak_macs]

    def get_kexs(self, allow_weak_kex):
        if allow_weak_kex:
            weak_kex = 'weak'
        else:
            weak_kex = 'default'

        default = 'diffie-hellman-group-exchange-sha256'
        weak = (default + ',diffie-hellman-group14-sha1,'
                'diffie-hellman-group-exchange-sha1,'
                'diffie-hellman-group1-sha1')
        kex = {'default': default,
               'weak': weak}

        default = ('curve25519-sha256@libssh.org,'
                   'diffie-hellman-group-exchange-sha256')
        weak = (default + ',diffie-hellman-group14-sha1,'
                'diffie-hellman-group-exchange-sha1,'
                'diffie-hellman-group1-sha1')
        kex_66 = {'default': default,
                  'weak': weak}

        # Use newer kex on Ubuntu Trusty and above
        _release = lsb_release()['DISTRIB_CODENAME'].lower()
        if CompareHostReleases(_release) >= 'trusty':
            log('Detected Ubuntu 14.04 or newer, using new key exchange '
                'algorithms', level=DEBUG)
            kex = kex_66

        return kex[weak_kex]

    def get_ciphers(self, cbc_required):
        if cbc_required:
            weak_ciphers = 'weak'
        else:
            weak_ciphers = 'default'

        default = 'aes256-ctr,aes192-ctr,aes128-ctr'
        cipher = {'default': default,
                  'weak': default + 'aes256-cbc,aes192-cbc,aes128-cbc'}

        default = ('chacha20-poly1305@openssh.com,aes256-gcm@openssh.com,'
                   'aes128-gcm@openssh.com,aes256-ctr,aes192-ctr,aes128-ctr')
        ciphers_66 = {'default': default,
                      'weak': default + ',aes256-cbc,aes192-cbc,aes128-cbc'}

        # Use newer ciphers on ubuntu Trusty and above
        _release = lsb_release()['DISTRIB_CODENAME'].lower()
        if CompareHostReleases(_release) >= 'trusty':
            log('Detected Ubuntu 14.04 or newer, using new ciphers',
                level=DEBUG)
            cipher = ciphers_66

        return cipher[weak_ciphers]

    def get_listening(self, listen=['0.0.0.0']):
        """Returns a list of addresses SSH can list on

        Turns input into a sensible list of IPs SSH can listen on. Input
        must be a python list of interface names, IPs and/or CIDRs.

        :param listen: list of IPs, CIDRs, interface names

        :returns: list of IPs available on the host
        """
        if listen == ['0.0.0.0']:
            return listen

        value = []
        for network in listen:
            try:
                ip = get_address_in_network(network=network, fatal=True)
            except ValueError:
                if is_ip(network):
                    ip = network
                else:
                    try:
                        ip = get_iface_addr(iface=network, fatal=False)[0]
                    except IndexError:
                        continue
            value.append(ip)
        if value == []:
            return ['0.0.0.0']
        return value

    def __call__(self):
        settings = utils.get_settings('ssh')
        if settings['common']['network_ipv6_enable']:
            addr_family = 'any'
        else:
            addr_family = 'inet'

        ctxt = {
            'addr_family': addr_family,
            'remote_hosts': settings['common']['remote_hosts'],
            'password_auth_allowed':
            settings['client']['password_authentication'],
            'ports': settings['common']['ports'],
            'ciphers': self.get_ciphers(settings['client']['cbc_required']),
            'macs': self.get_macs(settings['client']['weak_hmac']),
            'kexs': self.get_kexs(settings['client']['weak_kex']),
            'roaming': settings['client']['roaming'],
        }
        return ctxt


class SSHConfig(TemplatedFile):
    def __init__(self):
        path = '/etc/ssh/ssh_config'
        super(SSHConfig, self).__init__(path=path,
                                        template_dir=TEMPLATES_DIR,
                                        context=SSHConfigContext(),
                                        user='root',
                                        group='root',
                                        mode=0o0644)

    def pre_write(self):
        settings = utils.get_settings('ssh')
        apt_update(fatal=True)
        apt_install(settings['client']['package'])
        if not os.path.exists('/etc/ssh'):
            os.makedir('/etc/ssh')
            # NOTE: don't recurse
            utils.ensure_permissions('/etc/ssh', 'root', 'root', 0o0755,
                                     maxdepth=0)

    def post_write(self):
        # NOTE: don't recurse
        utils.ensure_permissions('/etc/ssh', 'root', 'root', 0o0755,
                                 maxdepth=0)


class SSHDConfigContext(SSHConfigContext):

    type = 'server'

    def __call__(self):
        settings = utils.get_settings('ssh')
        if settings['common']['network_ipv6_enable']:
            addr_family = 'any'
        else:
            addr_family = 'inet'

        ctxt = {
            'ssh_ip': self.get_listening(settings['server']['listen_to']),
            'password_auth_allowed':
            settings['server']['password_authentication'],
            'ports': settings['common']['ports'],
            'addr_family': addr_family,
            'ciphers': self.get_ciphers(settings['server']['cbc_required']),
            'macs': self.get_macs(settings['server']['weak_hmac']),
            'kexs': self.get_kexs(settings['server']['weak_kex']),
            'host_key_files': settings['server']['host_key_files'],
            'allow_root_with_key': settings['server']['allow_root_with_key'],
            'password_authentication':
            settings['server']['password_authentication'],
            'use_priv_sep': settings['server']['use_privilege_separation'],
            'use_pam': settings['server']['use_pam'],
            'allow_x11_forwarding': settings['server']['allow_x11_forwarding'],
            'print_motd': settings['server']['print_motd'],
            'print_last_log': settings['server']['print_last_log'],
            'client_alive_interval':
            settings['server']['alive_interval'],
            'client_alive_count': settings['server']['alive_count'],
            'allow_tcp_forwarding': settings['server']['allow_tcp_forwarding'],
            'allow_agent_forwarding':
            settings['server']['allow_agent_forwarding'],
            'deny_users': settings['server']['deny_users'],
            'allow_users': settings['server']['allow_users'],
            'deny_groups': settings['server']['deny_groups'],
            'allow_groups': settings['server']['allow_groups'],
            'use_dns': settings['server']['use_dns'],
            'sftp_enable': settings['server']['sftp_enable'],
            'sftp_group': settings['server']['sftp_group'],
            'sftp_chroot': settings['server']['sftp_chroot'],
            'max_auth_tries': settings['server']['max_auth_tries'],
            'max_sessions': settings['server']['max_sessions'],
        }
        return ctxt


class SSHDConfig(TemplatedFile):
    def __init__(self):
        path = '/etc/ssh/sshd_config'
        super(SSHDConfig, self).__init__(path=path,
                                         template_dir=TEMPLATES_DIR,
                                         context=SSHDConfigContext(),
                                         user='root',
                                         group='root',
                                         mode=0o0600,
                                         service_actions=[{'service': 'ssh',
                                                           'actions':
                                                           ['restart']}])

    def pre_write(self):
        settings = utils.get_settings('ssh')
        apt_update(fatal=True)
        apt_install(settings['server']['package'])
        if not os.path.exists('/etc/ssh'):
            os.makedir('/etc/ssh')
            # NOTE: don't recurse
            utils.ensure_permissions('/etc/ssh', 'root', 'root', 0o0755,
                                     maxdepth=0)

    def post_write(self):
        # NOTE: don't recurse
        utils.ensure_permissions('/etc/ssh', 'root', 'root', 0o0755,
                                 maxdepth=0)


class SSHConfigFileContentAudit(FileContentAudit):
    def __init__(self):
        self.path = '/etc/ssh/ssh_config'
        super(SSHConfigFileContentAudit, self).__init__(self.path, {})

    def is_compliant(self, *args, **kwargs):
        self.pass_cases = []
        self.fail_cases = []
        settings = utils.get_settings('ssh')

        _release = lsb_release()['DISTRIB_CODENAME'].lower()
        if CompareHostReleases(_release) >= 'trusty':
            if not settings['server']['weak_hmac']:
                self.pass_cases.append(r'^MACs.+,hmac-ripemd160$')
            else:
                self.pass_cases.append(r'^MACs.+,hmac-sha1$')

            if settings['server']['weak_kex']:
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha256[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group14-sha1[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha1[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group1-sha1[,\s]?')  # noqa
            else:
                self.pass_cases.append(r'^KexAlgorithms.+,diffie-hellman-group-exchange-sha256$')  # noqa
                self.fail_cases.append(r'^KexAlgorithms.*diffie-hellman-group14-sha1[,\s]?')  # noqa

            if settings['server']['cbc_required']:
                self.pass_cases.append(r'^Ciphers\s.*-cbc[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes128-ctr[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes192-ctr[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes256-ctr[,\s]?')
            else:
                self.fail_cases.append(r'^Ciphers\s.*-cbc[,\s]?')
                self.pass_cases.append(r'^Ciphers\schacha20-poly1305@openssh.com,.+')  # noqa
                self.pass_cases.append(r'^Ciphers\s.*aes128-ctr$')
                self.pass_cases.append(r'^Ciphers\s.*aes192-ctr[,\s]?')
                self.pass_cases.append(r'^Ciphers\s.*aes256-ctr[,\s]?')
        else:
            if not settings['client']['weak_hmac']:
                self.fail_cases.append(r'^MACs.+,hmac-sha1$')
            else:
                self.pass_cases.append(r'^MACs.+,hmac-sha1$')

            if settings['client']['weak_kex']:
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha256[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group14-sha1[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha1[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group1-sha1[,\s]?')  # noqa
            else:
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha256$')  # noqa
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group14-sha1[,\s]?')  # noqa
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha1[,\s]?')  # noqa
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group1-sha1[,\s]?')  # noqa

            if settings['client']['cbc_required']:
                self.pass_cases.append(r'^Ciphers\s.*-cbc[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes128-ctr[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes192-ctr[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes256-ctr[,\s]?')
            else:
                self.fail_cases.append(r'^Ciphers\s.*-cbc[,\s]?')
                self.pass_cases.append(r'^Ciphers\s.*aes128-ctr[,\s]?')
                self.pass_cases.append(r'^Ciphers\s.*aes192-ctr[,\s]?')
                self.pass_cases.append(r'^Ciphers\s.*aes256-ctr[,\s]?')

        if settings['client']['roaming']:
            self.pass_cases.append(r'^UseRoaming yes$')
        else:
            self.fail_cases.append(r'^UseRoaming yes$')

        return super(SSHConfigFileContentAudit, self).is_compliant(*args,
                                                                   **kwargs)


class SSHDConfigFileContentAudit(FileContentAudit):
    def __init__(self):
        self.path = '/etc/ssh/sshd_config'
        super(SSHDConfigFileContentAudit, self).__init__(self.path, {})

    def is_compliant(self, *args, **kwargs):
        self.pass_cases = []
        self.fail_cases = []
        settings = utils.get_settings('ssh')

        _release = lsb_release()['DISTRIB_CODENAME'].lower()
        if CompareHostReleases(_release) >= 'trusty':
            if not settings['server']['weak_hmac']:
                self.pass_cases.append(r'^MACs.+,hmac-ripemd160$')
            else:
                self.pass_cases.append(r'^MACs.+,hmac-sha1$')

            if settings['server']['weak_kex']:
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha256[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group14-sha1[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha1[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group1-sha1[,\s]?')  # noqa
            else:
                self.pass_cases.append(r'^KexAlgorithms.+,diffie-hellman-group-exchange-sha256$')  # noqa
                self.fail_cases.append(r'^KexAlgorithms.*diffie-hellman-group14-sha1[,\s]?')  # noqa

            if settings['server']['cbc_required']:
                self.pass_cases.append(r'^Ciphers\s.*-cbc[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes128-ctr[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes192-ctr[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes256-ctr[,\s]?')
            else:
                self.fail_cases.append(r'^Ciphers\s.*-cbc[,\s]?')
                self.pass_cases.append(r'^Ciphers\schacha20-poly1305@openssh.com,.+')  # noqa
                self.pass_cases.append(r'^Ciphers\s.*aes128-ctr$')
                self.pass_cases.append(r'^Ciphers\s.*aes192-ctr[,\s]?')
                self.pass_cases.append(r'^Ciphers\s.*aes256-ctr[,\s]?')
        else:
            if not settings['server']['weak_hmac']:
                self.pass_cases.append(r'^MACs.+,hmac-ripemd160$')
            else:
                self.pass_cases.append(r'^MACs.+,hmac-sha1$')

            if settings['server']['weak_kex']:
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha256[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group14-sha1[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha1[,\s]?')  # noqa
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group1-sha1[,\s]?')  # noqa
            else:
                self.pass_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha256$')  # noqa
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group14-sha1[,\s]?')  # noqa
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group-exchange-sha1[,\s]?')  # noqa
                self.fail_cases.append(r'^KexAlgorithms\sdiffie-hellman-group1-sha1[,\s]?')  # noqa

            if settings['server']['cbc_required']:
                self.pass_cases.append(r'^Ciphers\s.*-cbc[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes128-ctr[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes192-ctr[,\s]?')
                self.fail_cases.append(r'^Ciphers\s.*aes256-ctr[,\s]?')
            else:
                self.fail_cases.append(r'^Ciphers\s.*-cbc[,\s]?')
                self.pass_cases.append(r'^Ciphers\s.*aes128-ctr[,\s]?')
                self.pass_cases.append(r'^Ciphers\s.*aes192-ctr[,\s]?')
                self.pass_cases.append(r'^Ciphers\s.*aes256-ctr[,\s]?')

        if settings['server']['sftp_enable']:
            self.pass_cases.append(r'^Subsystem\ssftp')
        else:
            self.fail_cases.append(r'^Subsystem\ssftp')

        return super(SSHDConfigFileContentAudit, self).is_compliant(*args,
                                                                    **kwargs)



# Copyright 2014-2021 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Various utilities for dealing with Neutron and the renaming from Quantum.

import six
from subprocess import check_output

from charmhelpers.core.hookenv import (
    config,
    log,
    ERROR,
)

from charmhelpers.contrib.openstack.utils import (
    os_release,
    CompareOpenStackReleases,
)


def headers_package():
    """Ensures correct linux-headers for running kernel are installed,
    for building DKMS package"""
    kver = check_output(['uname', '-r']).decode('UTF-8').strip()
    return 'linux-headers-%s' % kver


QUANTUM_CONF_DIR = '/etc/quantum'


def kernel_version():
    """ Retrieve the current major kernel version as a tuple e.g. (3, 13) """
    kver = check_output(['uname', '-r']).decode('UTF-8').strip()
    kver = kver.split('.')
    return (int(kver[0]), int(kver[1]))


def determine_dkms_package():
    """ Determine which DKMS package should be used based on kernel version """
    # NOTE: 3.13 kernels have support for GRE and VXLAN native
    if kernel_version() >= (3, 13):
        return []
    else:
        return [headers_package(), 'openvswitch-datapath-dkms']


# legacy


def quantum_plugins():
    return {
        'ovs': {
            'config': '/etc/quantum/plugins/openvswitch/'
                      'ovs_quantum_plugin.ini',
            'driver': 'quantum.plugins.openvswitch.ovs_quantum_plugin.'
                      'OVSQuantumPluginV2',
            'contexts': [],
            'services': ['quantum-plugin-openvswitch-agent'],
            'packages': [determine_dkms_package(),
                         ['quantum-plugin-openvswitch-agent']],
            'server_packages': ['quantum-server',
                                'quantum-plugin-openvswitch'],
            'server_services': ['quantum-server']
        },
        'nvp': {
            'config': '/etc/quantum/plugins/nicira/nvp.ini',
            'driver': 'quantum.plugins.nicira.nicira_nvp_plugin.'
                      'QuantumPlugin.NvpPluginV2',
            'contexts': [],
            'services': [],
            'packages': [],
            'server_packages': ['quantum-server',
                                'quantum-plugin-nicira'],
            'server_services': ['quantum-server']
        }
    }


NEUTRON_CONF_DIR = '/etc/neutron'


def neutron_plugins():
    release = os_release('nova-common')
    plugins = {
        'ovs': {
            'config': '/etc/neutron/plugins/openvswitch/'
                      'ovs_neutron_plugin.ini',
            'driver': 'neutron.plugins.openvswitch.ovs_neutron_plugin.'
                      'OVSNeutronPluginV2',
            'contexts': [],
            'services': ['neutron-plugin-openvswitch-agent'],
            'packages': [determine_dkms_package(),
                         ['neutron-plugin-openvswitch-agent']],
            'server_packages': ['neutron-server',
                                'neutron-plugin-openvswitch'],
            'server_services': ['neutron-server']
        },
        'nvp': {
            'config': '/etc/neutron/plugins/nicira/nvp.ini',
            'driver': 'neutron.plugins.nicira.nicira_nvp_plugin.'
                      'NeutronPlugin.NvpPluginV2',
            'contexts': [],
            'services': [],
            'packages': [],
            'server_packages': ['neutron-server',
                                'neutron-plugin-nicira'],
            'server_services': ['neutron-server']
        },
        'nsx': {
            'config': '/etc/neutron/plugins/vmware/nsx.ini',
            'driver': 'vmware',
            'contexts': [],
            'services': [],
            'packages': [],
            'server_packages': ['neutron-server',
                                'neutron-plugin-vmware'],
            'server_services': ['neutron-server']
        },
        'n1kv': {
            'config': '/etc/neutron/plugins/cisco/cisco_plugins.ini',
            'driver': 'neutron.plugins.cisco.network_plugin.PluginV2',
            'contexts': [],
            'services': [],
            'packages': [determine_dkms_package(),
                         ['neutron-plugin-cisco']],
            'server_packages': ['neutron-server',
                                'neutron-plugin-cisco'],
            'server_services': ['neutron-server']
        },
        'Calico': {
            'config': '/etc/neutron/plugins/ml2/ml2_conf.ini',
            'driver': 'neutron.plugins.ml2.plugin.Ml2Plugin',
            'contexts': [],
            'services': ['calico-felix',
                         'bird',
                         'neutron-dhcp-agent',
                         'nova-api-metadata',
                         'etcd'],
            'packages': [determine_dkms_package(),
                         ['calico-compute',
                          'bird',
                          'neutron-dhcp-agent',
                          'nova-api-metadata',
                          'etcd']],
            'server_packages': ['neutron-server', 'calico-control', 'etcd'],
            'server_services': ['neutron-server', 'etcd']
        },
        'vsp': {
            'config': '/etc/neutron/plugins/nuage/nuage_plugin.ini',
            'driver': 'neutron.plugins.nuage.plugin.NuagePlugin',
            'contexts': [],
            'services': [],
            'packages': [],
            'server_packages': ['neutron-server', 'neutron-plugin-nuage'],
            'server_services': ['neutron-server']
        },
        'plumgrid': {
            'config': '/etc/neutron/plugins/plumgrid/plumgrid.ini',
            'driver': ('neutron.plugins.plumgrid.plumgrid_plugin'
                       '.plumgrid_plugin.NeutronPluginPLUMgridV2'),
            'contexts': [],
            'services': [],
            'packages': ['plumgrid-lxc',
                         'iovisor-dkms'],
            'server_packages': ['neutron-server',
                                'neutron-plugin-plumgrid'],
            'server_services': ['neutron-server']
        },
        'midonet': {
            'config': '/etc/neutron/plugins/midonet/midonet.ini',
            'driver': 'midonet.neutron.plugin.MidonetPluginV2',
            'contexts': [],
            'services': [],
            'packages': [determine_dkms_package()],
            'server_packages': ['neutron-server',
                                'python-neutron-plugin-midonet'],
            'server_services': ['neutron-server']
        }
    }
    if CompareOpenStackReleases(release) >= 'icehouse':
        # NOTE: patch in ml2 plugin for icehouse onwards
        plugins['ovs']['config'] = '/etc/neutron/plugins/ml2/ml2_conf.ini'
        plugins['ovs']['driver'] = 'neutron.plugins.ml2.plugin.Ml2Plugin'
        plugins['ovs']['server_packages'] = ['neutron-server',
                                             'neutron-plugin-ml2']
        # NOTE: patch in vmware renames nvp->nsx for icehouse onwards
        plugins['nvp'] = plugins['nsx']
    if CompareOpenStackReleases(release) >= 'kilo':
        plugins['midonet']['driver'] = (
            'neutron.plugins.midonet.plugin.MidonetPluginV2')
    if CompareOpenStackReleases(release) >= 'liberty':
        plugins['midonet']['driver'] = (
            'midonet.neutron.plugin_v1.MidonetPluginV2')
        plugins['midonet']['server_packages'].remove(
            'python-neutron-plugin-midonet')
        plugins['midonet']['server_packages'].append(
            'python-networking-midonet')
        plugins['plumgrid']['driver'] = (
            'networking_plumgrid.neutron.plugins'
            '.plugin.NeutronPluginPLUMgridV2')
        plugins['plumgrid']['server_packages'].remove(
            'neutron-plugin-plumgrid')
    if CompareOpenStackReleases(release) >= 'mitaka':
        plugins['nsx']['server_packages'].remove('neutron-plugin-vmware')
        plugins['nsx']['server_packages'].append('python-vmware-nsx')
        plugins['nsx']['config'] = '/etc/neutron/nsx.ini'
        plugins['vsp']['driver'] = (
            'nuage_neutron.plugins.nuage.plugin.NuagePlugin')
    if CompareOpenStackReleases(release) >= 'newton':
        plugins['vsp']['config'] = '/etc/neutron/plugins/ml2/ml2_conf.ini'
        plugins['vsp']['driver'] = 'neutron.plugins.ml2.plugin.Ml2Plugin'
        plugins['vsp']['server_packages'] = ['neutron-server',
                                             'neutron-plugin-ml2']
    return plugins


def neutron_plugin_attribute(plugin, attr, net_manager=None):
    manager = net_manager or network_manager()
    if manager == 'quantum':
        plugins = quantum_plugins()
    elif manager == 'neutron':
        plugins = neutron_plugins()
    else:
        log("Network manager '%s' does not support plugins." % (manager),
            level=ERROR)
        raise Exception

    try:
        _plugin = plugins[plugin]
    except KeyError:
        log('Unrecognised plugin for %s: %s' % (manager, plugin), level=ERROR)
        raise Exception

    try:
        return _plugin[attr]
    except KeyError:
        return None


def network_manager():
    '''
    Deals with the renaming of Quantum to Neutron in H and any situations
    that require compatibility (eg, deploying H with network-manager=quantum,
    upgrading from G).
    '''
    release = os_release('nova-common')
    manager = config('network-manager').lower()

    if manager not in ['quantum', 'neutron']:
        return manager

    if release in ['essex']:
        # E does not support neutron
        log('Neutron networking not supported in Essex.', level=ERROR)
        raise Exception
    elif release in ['folsom', 'grizzly']:
        # neutron is named quantum in F and G
        return 'quantum'
    else:
        # ensure accurate naming for all releases post-H
        return 'neutron'


def parse_mappings(mappings, key_rvalue=False):
    """By default mappings are lvalue keyed.

    If key_rvalue is True, the mapping will be reversed to allow multiple
    configs for the same lvalue.
    """
    parsed = {}
    if mappings:
        mappings = mappings.split()
        for m in mappings:
            p = m.partition(':')

            if key_rvalue:
                key_index = 2
                val_index = 0
                # if there is no rvalue skip to next
                if not p[1]:
                    continue
            else:
                key_index = 0
                val_index = 2

            key = p[key_index].strip()
            parsed[key] = p[val_index].strip()

    return parsed


def parse_bridge_mappings(mappings):
    """Parse bridge mappings.

    Mappings must be a space-delimited list of provider:bridge mappings.

    Returns dict of the form {provider:bridge}.
    """
    return parse_mappings(mappings)


def parse_data_port_mappings(mappings, default_bridge='br-data'):
    """Parse data port mappings.

    Mappings must be a space-delimited list of bridge:port.

    Returns dict of the form {port:bridge} where ports may be mac addresses or
    interface names.
    """

    # NOTE(dosaboy): we use rvalue for key to allow multiple values to be
    # proposed for <port> since it may be a mac address which will differ
    # across units this allowing first-known-good to be chosen.
    _mappings = parse_mappings(mappings, key_rvalue=True)
    if not _mappings or list(_mappings.values()) == ['']:
        if not mappings:
            return {}

        # For backwards-compatibility we need to support port-only provided in
        # config.
        _mappings = {mappings.split()[0]: default_bridge}

    ports = _mappings.keys()
    if len(set(ports)) != len(ports):
        raise Exception("It is not allowed to have the same port configured "
                        "on more than one bridge")

    return _mappings


def parse_vlan_range_mappings(mappings):
    """Parse vlan range mappings.

    Mappings must be a space-delimited list of provider:start:end mappings.

    The start:end range is optional and may be omitted.

    Returns dict of the form {provider: (start, end)}.
    """
    _mappings = parse_mappings(mappings)
    if not _mappings:
        return {}

    mappings = {}
    for p, r in six.iteritems(_mappings):
        mappings[p] = tuple(r.split(':'))

    return mappings



# Copyright 2017 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json

from charmhelpers.core.hookenv import log

stats_intervals = ['stats_day', 'stats_five_minute',
                   'stats_hour', 'stats_total']

SYSFS = '/sys'


class Bcache(object):
    """Bcache behaviour
    """

    def __init__(self, cachepath):
        self.cachepath = cachepath

    @classmethod
    def fromdevice(cls, devname):
        return cls('{}/block/{}/bcache'.format(SYSFS, devname))

    def __str__(self):
        return self.cachepath

    def get_stats(self, interval):
        """Get cache stats
        """
        intervaldir = 'stats_{}'.format(interval)
        path = "{}/{}".format(self.cachepath, intervaldir)
        out = dict()
        for elem in os.listdir(path):
            out[elem] = open('{}/{}'.format(path, elem)).read().strip()
        return out


def get_bcache_fs():
    """Return all cache sets
    """
    cachesetroot = "{}/fs/bcache".format(SYSFS)
    try:
        dirs = os.listdir(cachesetroot)
    except OSError:
        log("No bcache fs found")
        return []
    cacheset = set([Bcache('{}/{}'.format(cachesetroot, d)) for d in dirs if not d.startswith('register')])
    return cacheset


def get_stats_action(cachespec, interval):
    """Action for getting bcache statistics for a given cachespec.
    Cachespec can either be a device name, eg. 'sdb', which will retrieve
    cache stats for the given device, or 'global', which will retrieve stats
    for all cachesets
    """
    if cachespec == 'global':
        caches = get_bcache_fs()
    else:
        caches = [Bcache.fromdevice(cachespec)]
    res = dict((c.cachepath, c.get_stats(interval)) for c in caches)
    return json.dumps(res, indent=4, separators=(',', ': '))



# -*- coding: utf-8 -*-

# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
from charmhelpers.core import fstab
from charmhelpers.core import sysctl
from charmhelpers.core.host import (
    add_group,
    add_user_to_group,
    fstab_mount,
    mkdir,
)
from charmhelpers.core.strutils import bytes_from_string
from subprocess import check_output


def hugepage_support(user, group='hugetlb', nr_hugepages=256,
                     max_map_count=65536, mnt_point='/run/hugepages/kvm',
                     pagesize='2MB', mount=True, set_shmmax=False):
    """Enable hugepages on system.

    Args:
    user (str)  -- Username to allow access to hugepages to
    group (str) -- Group name to own hugepages
    nr_hugepages (int) -- Number of pages to reserve
    max_map_count (int) -- Number of Virtual Memory Areas a process can own
    mnt_point (str) -- Directory to mount hugepages on
    pagesize (str) -- Size of hugepages
    mount (bool) -- Whether to Mount hugepages
    """
    group_info = add_group(group)
    gid = group_info.gr_gid
    add_user_to_group(user, group)
    if max_map_count < 2 * nr_hugepages:
        max_map_count = 2 * nr_hugepages
    sysctl_settings = {
        'vm.nr_hugepages': nr_hugepages,
        'vm.max_map_count': max_map_count,
        'vm.hugetlb_shm_group': gid,
    }
    if set_shmmax:
        shmmax_current = int(check_output(['sysctl', '-n', 'kernel.shmmax']))
        shmmax_minsize = bytes_from_string(pagesize) * nr_hugepages
        if shmmax_minsize > shmmax_current:
            sysctl_settings['kernel.shmmax'] = shmmax_minsize
    sysctl.create(yaml.dump(sysctl_settings), '/etc/sysctl.d/10-hugepage.conf')
    mkdir(mnt_point, owner='root', group='root', perms=0o755, force=False)
    lfstab = fstab.Fstab()
    fstab_entry = lfstab.get_entry_by_attr('mountpoint', mnt_point)
    if fstab_entry:
        lfstab.remove_entry(fstab_entry)
    entry = lfstab.Entry('nodev', mnt_point, 'hugetlbfs',
                         'mode=1770,gid={},pagesize={}'.format(gid, pagesize), 0, 0)
    lfstab.add_entry(entry)
    if mount:
        fstab_mount(mnt_point)



import os
import mock
import json
import unittest
import sys
import shutil
import tempfile

from collections import OrderedDict

sys.modules['MySQLdb'] = mock.Mock()
from charmhelpers.contrib.database import mysql  # noqa


class MysqlTests(unittest.TestCase):
    def setUp(self):
        super(MysqlTests, self).setUp()

    def test_connect_host_defined(self):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        with mock.patch.object(mysql, 'log'):
            helper.connect(user='user', password='password', host='1.1.1.1')
        mysql.MySQLdb.connect.assert_called_with(
            passwd='password', host='1.1.1.1', user='user', connect_timeout=30)

    def test_connect_host_not_defined(self):
        helper = mysql.MySQLHelper('foo', 'bar')
        with mock.patch.object(mysql, 'log'):
            helper.connect(user='user', password='password')
        mysql.MySQLdb.connect.assert_called_with(
            passwd='password', host='localhost', user='user',
            connect_timeout=30)

    def test_connect_port_defined(self):
        helper = mysql.MySQLHelper('foo', 'bar')
        with mock.patch.object(mysql, 'log'):
            helper.connect(user='user', password='password', port=3316)
        mysql.MySQLdb.connect.assert_called_with(
            passwd='password', host='localhost', user='user', port=3316,
            connect_timeout=30)

    def test_connect_new_default_timeout(self):
        helper = mysql.MySQLHelper('foo', 'bar', connect_timeout=10)
        with mock.patch.object(mysql, 'log'):
            helper.connect(user='user', password='password', port=3316)
        mysql.MySQLdb.connect.assert_called_with(
            passwd='password', host='localhost', user='user', port=3316,
            connect_timeout=10)

    def test_connect_new_default_override(self):
        helper = mysql.MySQLHelper('foo', 'bar', connect_timeout=10)
        with mock.patch.object(mysql, 'log'):
            helper.connect(user='user', password='password', port=3316,
                           connect_timeout=20)
        mysql.MySQLdb.connect.assert_called_with(
            passwd='password', host='localhost', user='user', port=3316,
            connect_timeout=20)

    @mock.patch.object(mysql.MySQLHelper, 'normalize_address')
    @mock.patch.object(mysql.MySQLHelper, 'get_mysql_password')
    @mock.patch.object(mysql.MySQLHelper, 'grant_exists')
    @mock.patch.object(mysql, 'relation_get')
    @mock.patch.object(mysql, 'related_units')
    @mock.patch.object(mysql, 'log')
    def test_get_allowed_units(self, mock_log, mock_related_units,
                               mock_relation_get,
                               mock_grant_exists,
                               mock_get_password,
                               mock_normalize_address):

        # echo
        mock_normalize_address.side_effect = lambda addr: addr

        def mock_rel_get(unit, rid):
            if unit == 'unit/0':
                # Non-prefixed settings
                d = {'private-address': '10.0.0.1',
                     'hostname': 'hostA'}
            elif unit == 'unit/1':
                # Containing prefixed settings
                d = {'private-address': '10.0.0.2',
                     'dbA_hostname': json.dumps(['10.0.0.2', '2001:db8:1::2'])}
            elif unit == 'unit/2':
                # No hostname
                d = {'private-address': '10.0.0.3'}
            elif unit == 'unit/3':
                # Prefixed hostname
                d = {'private-address': '10.0.0.4',
                     'PRE_hostname': json.dumps(['10.0.0.4', '2001:db8:1::4'])}
            return d

        mock_relation_get.side_effect = mock_rel_get
        mock_related_units.return_value = ['unit/0', 'unit/1', 'unit/2']

        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        units = helper.get_allowed_units('dbA', 'userA')

        calls = [mock.call('dbA', 'userA', 'hostA'),
                 mock.call('dbA', 'userA', '10.0.0.2'),
                 mock.call('dbA', 'userA', '2001:db8:1::2'),
                 mock.call('dbA', 'userA', '10.0.0.3')]

        helper.grant_exists.assert_has_calls(calls, any_order=True)
        self.assertEqual(units, set(['unit/0', 'unit/1', 'unit/2']))

        # With prefix
        calls = [mock.call('dbB', 'userB', 'hostA'),
                 mock.call('dbB', 'userB', '10.0.0.2'),
                 mock.call('dbB', 'userB', '10.0.0.3'),
                 mock.call('dbB', 'userB', '2001:db8:1::4'),
                 mock.call('dbB', 'userB', '10.0.0.4')]

        mock_related_units.return_value = [
            'unit/0', 'unit/1', 'unit/2', 'unit/3']
        units = helper.get_allowed_units('dbB', 'userB', prefix="PRE")
        helper.grant_exists.assert_has_calls(calls, any_order=True)
        self.assertEqual(units, set(['unit/0', 'unit/1', 'unit/2', 'unit/3']))

    @mock.patch('charmhelpers.contrib.network.ip.log',
                lambda *args, **kwargs: None)
    @mock.patch('charmhelpers.contrib.network.ip.ns_query')
    @mock.patch('charmhelpers.contrib.network.ip.socket')
    @mock.patch.object(mysql, 'unit_get')
    @mock.patch.object(mysql, 'config_get')
    @mock.patch.object(mysql, 'log')
    def test_normalize_address(self, mock_log, mock_config_get, mock_unit_get,
                               mock_socket, mock_ns_query):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        # prefer-ipv6
        mock_config_get.return_value = False
        # echo
        mock_socket.gethostbyname.side_effect = lambda addr: addr

        mock_unit_get.return_value = '10.0.0.1'
        out = helper.normalize_address('10.0.0.1')
        self.assertEqual('127.0.0.1', out)
        mock_config_get.assert_called_with('prefer-ipv6')

        mock_unit_get.return_value = '10.0.0.1'
        out = helper.normalize_address('10.0.0.2')
        self.assertEqual('10.0.0.2', out)
        mock_config_get.assert_called_with('prefer-ipv6')

        out = helper.normalize_address('2001:db8:1::1')
        self.assertEqual('2001:db8:1::1', out)
        mock_config_get.assert_called_with('prefer-ipv6')

        mock_socket.gethostbyname.side_effect = Exception
        mock_ns_query.return_value = None
        out = helper.normalize_address('unresolvable')
        self.assertEqual('unresolvable', out)
        mock_config_get.assert_called_with('prefer-ipv6')

        # prefer-ipv6
        mock_config_get.return_value = True
        mock_socket.gethostbyname.side_effect = 'other'
        out = helper.normalize_address('unresolvable')
        self.assertEqual('unresolvable', out)
        mock_config_get.assert_called_with('prefer-ipv6')

    def test_passwd_keys(self):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        self.assertEqual(list(helper.passwd_keys(None)), ['mysql.passwd'])
        self.assertEqual(list(helper.passwd_keys('auser')),
                         ['mysql-auser.passwd', 'auser.passwd'])

    @mock.patch.object(mysql.MySQLHelper, 'migrate_passwords_to_leader_storage')
    @mock.patch.object(mysql.MySQLHelper, 'get_mysql_password_on_disk')
    @mock.patch.object(mysql, 'leader_get')
    def test_get_mysql_password_no_peer_passwd(self, mock_leader_get,
                                               mock_get_disk_pw,
                                               mock_migrate_pw):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        store = {}
        mock_leader_get.side_effect = lambda key: store.get(key)
        mock_get_disk_pw.return_value = "disk-passwd"
        self.assertEqual(helper.get_mysql_password(), "disk-passwd")
        self.assertTrue(mock_migrate_pw.called)

    @mock.patch.object(mysql.MySQLHelper, 'migrate_passwords_to_leader_storage')
    @mock.patch.object(mysql.MySQLHelper, 'get_mysql_password_on_disk')
    @mock.patch.object(mysql, 'leader_get')
    def test_get_mysql_password_peer_passwd(self, mock_leader_get,
                                            mock_get_disk_pw, mock_migrate_pw):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        store = {'mysql-userA.passwd': 'passwdA'}
        mock_leader_get.side_effect = lambda key: store.get(key)
        mock_get_disk_pw.return_value = "disk-passwd"
        self.assertEqual(helper.get_mysql_password(username='userA'),
                         "passwdA")
        self.assertTrue(mock_migrate_pw.called)

    @mock.patch.object(mysql.MySQLHelper, 'migrate_passwords_to_leader_storage')
    @mock.patch.object(mysql.MySQLHelper, 'get_mysql_password_on_disk')
    @mock.patch.object(mysql, 'leader_get')
    def test_get_mysql_password_peer_passwd_legacy(self, mock_leader_get,
                                                   mock_get_disk_pw,
                                                   mock_migrate_pw):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        store = {'userA.passwd': 'passwdA'}
        mock_leader_get.side_effect = lambda key: store.get(key)
        mock_get_disk_pw.return_value = "disk-passwd"
        self.assertEqual(helper.get_mysql_password(username='userA'),
                         "passwdA")
        self.assertTrue(mock_migrate_pw.called)

    @mock.patch.object(mysql.MySQLHelper, 'migrate_passwords_to_leader_storage')
    @mock.patch.object(mysql.MySQLHelper, 'get_mysql_password_on_disk')
    @mock.patch.object(mysql, 'leader_get')
    def test_get_mysql_password_peer_passwd_all(self, mock_leader_get,
                                                mock_get_disk_pw,
                                                mock_migrate_pw):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        # Add * so we can identify that the new format key takes precedence
        # if found.
        store = {'mysql-userA.passwd': 'passwdA',
                 'userA.passwd': 'passwdA*'}
        mock_leader_get.side_effect = lambda key: store.get(key)
        mock_get_disk_pw.return_value = "disk-passwd"
        self.assertEqual(helper.get_mysql_password(username='userA'),
                         "passwdA")
        self.assertTrue(mock_migrate_pw.called)

    @mock.patch.object(mysql.MySQLHelper, 'set_mysql_password')
    def test_set_mysql_root_password(self, mock_set_passwd):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        helper.set_mysql_root_password(password='1234')
        mock_set_passwd.assert_called_with(
            'root',
            '1234',
            current_password=None)

    @mock.patch.object(mysql.MySQLHelper, 'set_mysql_password')
    def test_set_mysql_root_password_cur_passwd(self, mock_set_passwd):
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        helper.set_mysql_root_password(password='1234', current_password='abc')
        mock_set_passwd.assert_called_with(
            'root',
            '1234',
            current_password='abc')

    @mock.patch.object(mysql, 'log', lambda *args, **kwargs: None)
    @mock.patch.object(mysql, 'is_leader')
    @mock.patch.object(mysql, 'leader_get')
    @mock.patch.object(mysql, 'leader_set')
    @mock.patch.object(mysql, 'CompareHostReleases')
    @mock.patch.object(mysql.MySQLHelper, 'get_mysql_password')
    @mock.patch.object(mysql.MySQLHelper, 'connect')
    def test_set_mysql_password(self, mock_connect, mock_get_passwd,
                                mock_compare_releases, mock_leader_set,
                                mock_leader_get, mock_is_leader):
        mock_connection = mock.MagicMock()
        mock_cursor = mock.MagicMock()
        mock_connection.cursor.return_value = mock_cursor
        mock_get_passwd.return_value = 'asdf'
        mock_is_leader.return_value = True
        mock_leader_get.return_value = '1234'
        mock_compare_releases.return_value = 'artful'

        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        helper.connection = mock_connection

        helper.set_mysql_password(username='root', password='1234')

        mock_connect.assert_has_calls(
            [mock.call(user='root', password='asdf'),  # original password
             mock.call(user='root', password='1234')])  # new password
        mock_leader_get.assert_has_calls([mock.call('mysql.passwd')])
        mock_leader_set.assert_has_calls(
            [mock.call(settings={'mysql.passwd': '1234'})]
        )
        SQL_UPDATE_PASSWD = ("UPDATE mysql.user SET password = "
                             "PASSWORD( %s ) WHERE user = %s;")
        mock_cursor.assert_has_calls(
            [mock.call.execute(SQL_UPDATE_PASSWD, ('1234', 'root')),
             mock.call.execute('FLUSH PRIVILEGES;'),
             mock.call.close(),
             mock.call.execute('select 1;'),
             mock.call.close()]
        )
        mock_get_passwd.assert_called_once_with(None)

        # make sure for the non-leader leader-set is not called
        mock_is_leader.return_value = False
        mock_leader_set.reset_mock()
        mock_get_passwd.reset_mock()
        helper.set_mysql_password(username='root', password='1234')
        mock_leader_set.assert_not_called()
        mock_get_passwd.assert_called_once_with(None)

        mock_get_passwd.reset_mock()
        mock_compare_releases.return_value = 'bionic'
        helper.set_mysql_password(username='root', password='1234')
        SQL_UPDATE_PASSWD = ("UPDATE mysql.user SET "
                             "authentication_string = "
                             "PASSWORD( %s ) WHERE user = %s;")
        mock_cursor.assert_has_calls(
            [mock.call.execute(SQL_UPDATE_PASSWD, ('1234', 'root')),
             mock.call.execute('FLUSH PRIVILEGES;'),
             mock.call.close(),
             mock.call.execute('select 1;'),
             mock.call.close()]
        )
        mock_get_passwd.assert_called_once_with(None)

        # Test supplying the current password
        mock_is_leader.return_value = False
        mock_connect.reset_mock()
        mock_get_passwd.reset_mock()
        helper.set_mysql_password(
            username='root',
            password='1234',
            current_password='currpass')
        self.assertFalse(mock_get_passwd.called)
        mock_connect.assert_has_calls(
            [mock.call(user='root', password='currpass'),  # original password
             mock.call(user='root', password='1234')])  # new password

    @mock.patch.object(mysql, 'leader_get')
    @mock.patch.object(mysql, 'leader_set')
    @mock.patch.object(mysql.MySQLHelper, 'get_mysql_password')
    @mock.patch.object(mysql.MySQLHelper, 'connect')
    def test_set_mysql_password_fail_to_connect(self, mock_connect,
                                                mock_get_passwd,
                                                mock_leader_set,
                                                mock_leader_get):

        class FakeOperationalError(Exception):
            pass

        def fake_connect(*args, **kwargs):
            raise FakeOperationalError('foobar')

        mysql.MySQLdb.OperationalError = FakeOperationalError
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        mock_connect.side_effect = fake_connect
        self.assertRaises(mysql.MySQLSetPasswordError,
                          helper.set_mysql_password,
                          username='root', password='1234')

    @mock.patch.object(mysql, 'lsb_release')
    @mock.patch.object(mysql, 'leader_get')
    @mock.patch.object(mysql, 'leader_set')
    @mock.patch.object(mysql.MySQLHelper, 'get_mysql_password')
    @mock.patch.object(mysql.MySQLHelper, 'connect')
    def test_set_mysql_password_fail_to_connect2(self, mock_connect,
                                                 mock_get_passwd,
                                                 mock_leader_set,
                                                 mock_leader_get,
                                                 mock_lsb_release):

        class FakeOperationalError(Exception):
            def __str__(self):
                return 'some-error'

        operational_error = FakeOperationalError('foobar')

        def fake_connect(user, password):
            # fail for the new password
            if user == 'root' and password == '1234':
                raise operational_error
            else:
                return mock.MagicMock()

        mysql.MySQLdb.OperationalError = FakeOperationalError
        helper = mysql.MySQLHelper('foo', 'bar', host='hostA')
        helper.connection = mock.MagicMock()
        mock_connect.side_effect = fake_connect
        mock_lsb_release.return_value = {
            'DISTRIB_CODENAME': 'bionic',
        }
        with self.assertRaises(mysql.MySQLSetPasswordError) as cm:
            helper.set_mysql_password(username='root', password='1234')

        ex = cm.exception
        self.assertEqual(ex.args[0], 'Cannot connect using new password: some-error')
        self.assertEqual(ex.args[1], operational_error)

    @mock.patch.object(mysql, 'is_leader')
    @mock.patch.object(mysql, 'leader_set')
    def test_migrate_passwords_to_leader_storage(self, mock_leader_set,
                                                 mock_is_leader):
        files = {'mysql.passwd': '1',
                 'userA.passwd': '2',
                 'mysql-userA.passwd': '3'}
        store = {}

        def _store(settings):
            store.update(settings)

        mock_is_leader.return_value = True

        tmpdir = tempfile.mkdtemp('charm-helpers-unit-tests')
        try:
            root_tmplt = "%s/mysql.passwd" % (tmpdir)
            helper = mysql.MySQLHelper(root_tmplt, None, host='hostA')
            for f in files:
                with open(os.path.join(tmpdir, f), 'w') as fd:
                    fd.write(files[f])

            mock_leader_set.side_effect = _store
            helper.migrate_passwords_to_leader_storage()

            calls = [mock.call(settings={'mysql.passwd': '1'}),
                     mock.call(settings={'userA.passwd': '2'}),
                     mock.call(settings={'mysql-userA.passwd': '3'})]

            mock_leader_set.assert_has_calls(calls,
                                             any_order=True)
        finally:
            shutil.rmtree(tmpdir)

        # Note that legacy key/val is NOT overwritten
        self.assertEqual(store, {'mysql.passwd': '1',
                                 'userA.passwd': '2',
                                 'mysql-userA.passwd': '3'})

    @mock.patch.object(mysql, 'log', lambda *args, **kwargs: None)
    @mock.patch.object(mysql, 'is_leader')
    @mock.patch.object(mysql, 'leader_set')
    def test_migrate_passwords_to_leader_storage_not_leader(self, mock_leader_set,
                                                            mock_is_leader):
        mock_is_leader.return_value = False
        tmpdir = tempfile.mkdtemp('charm-helpers-unit-tests')
        try:
            root_tmplt = "%s/mysql.passwd" % (tmpdir)
            helper = mysql.MySQLHelper(root_tmplt, None, host='hostA')
            helper.migrate_passwords_to_leader_storage()
        finally:
            shutil.rmtree(tmpdir)
        mock_leader_set.assert_not_called()


class MySQLConfigHelperTests(unittest.TestCase):

    def setUp(self):
        super(MySQLConfigHelperTests, self).setUp()
        self.config_data = {}
        self.config = mock.MagicMock()
        mysql.config_get = self.config
        self.config.side_effect = self._fake_config

    def _fake_config(self, key=None):
        if key:
            try:
                return self.config_data[key]
            except KeyError:
                return None
        else:
            return self.config_data

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_innodb_pool_fixed(self, log, mem):
        mem.return_value = "100G"
        self.config_data = {
            'innodb-buffer-pool-size': "50%",
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_buffer_pool_size(),
            helper.human_to_bytes("50G"))

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_innodb_pool_not_set(self, mog, mem):
        mem.return_value = "100G"
        self.config_data = {
            'innodb-buffer-pool-size': '',
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_buffer_pool_size(),
            helper.DEFAULT_INNODB_BUFFER_SIZE_MAX)

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_innodb_buffer_unset(self, mog, mem):
        mem.return_value = "100G"
        self.config_data = {
            'innodb-buffer-pool-size': None,
            'dataset-size': None,
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_buffer_pool_size(),
            helper.DEFAULT_INNODB_BUFFER_SIZE_MAX)

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_innodb_buffer_unset_small(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'innodb-buffer-pool-size': None,
            'dataset-size': None,
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_buffer_pool_size(),
            int(helper.human_to_bytes(mem.return_value) *
                helper.DEFAULT_INNODB_BUFFER_FACTOR))

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_innodb_dataset_size(self, mog, mem):
        mem.return_value = "100G"
        self.config_data = {
            'dataset-size': "10G",
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_buffer_pool_size(),
            int(helper.human_to_bytes("10G")))

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_tuning_level(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'tuning-level': 'safest',
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_flush_log_at_trx_commit(),
            1
        )

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_tuning_level_fast(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'tuning-level': 'fast',
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_flush_log_at_trx_commit(),
            2
        )

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_tuning_level_unsafe(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'tuning-level': 'unsafe',
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_flush_log_at_trx_commit(),
            0
        )

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_innodb_valid_values(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'innodb-change-buffering': 'all',
        }

        helper = mysql.MySQLConfigHelper()

        self.assertEqual(
            helper.get_innodb_change_buffering(),
            'all'
        )

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_get_innodb_invalid_values(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'innodb-change-buffering': 'invalid',
        }

        helper = mysql.MySQLConfigHelper()

        self.assertTrue(helper.get_innodb_change_buffering() is None)


class PerconaTests(unittest.TestCase):

    def setUp(self):
        super(PerconaTests, self).setUp()
        self.config_data = {}
        self.config = mock.MagicMock()
        mysql.config_get = self.config
        self.config.side_effect = self._fake_config

    def _fake_config(self, key=None):
        if key:
            try:
                return self.config_data[key]
            except KeyError:
                return None
        else:
            return self.config_data

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_innodb_pool_fixed(self, log, mem):
        mem.return_value = "100G"
        self.config_data = {
            'innodb-buffer-pool-size': "50%",
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(mysql_config.get('innodb_buffer_pool_size'),
                         helper.human_to_bytes("50G"))

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_innodb_pool_not_set(self, mog, mem):
        mem.return_value = "100G"
        self.config_data = {
            'innodb-buffer-pool-size': '',
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('innodb_buffer_pool_size'),
            helper.DEFAULT_INNODB_BUFFER_SIZE_MAX)

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_innodb_buffer_unset(self, mog, mem):
        mem.return_value = "100G"
        self.config_data = {
            'innodb-buffer-pool-size': None,
            'dataset-size': None,
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('innodb_buffer_pool_size'),
            helper.DEFAULT_INNODB_BUFFER_SIZE_MAX)

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_innodb_buffer_unset_small(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'innodb-buffer-pool-size': None,
            'dataset-size': None,
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('innodb_buffer_pool_size'),
            int(helper.human_to_bytes(mem.return_value) *
                helper.DEFAULT_INNODB_BUFFER_FACTOR))

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_innodb_dataset_size(self, mog, mem):
        mem.return_value = "100G"
        self.config_data = {
            'dataset-size': "10G",
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('innodb_buffer_pool_size'),
            int(helper.human_to_bytes("10G")))

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_wait_timeout(self, mog, mem):
        mem.return_value = "100G"

        timeout = 314
        self.config_data = {
            'wait-timeout': timeout,
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('wait_timeout'),
            timeout)

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_tuning_level(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'tuning-level': 'safest',
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('innodb_flush_log_at_trx_commit'),
            1
        )

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_tuning_level_fast(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'tuning-level': 'fast',
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('innodb_flush_log_at_trx_commit'),
            2
        )

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_tuning_level_unsafe(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'tuning-level': 'unsafe',
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('innodb_flush_log_at_trx_commit'),
            0
        )

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_innodb_valid_values(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'innodb-change-buffering': 'all',
            'innodb-io-capacity': 100,
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertEqual(
            mysql_config.get('innodb_change_buffering'),
            'all'
        )

        self.assertEqual(
            mysql_config.get('innodb_io_capacity'),
            100
        )

    @mock.patch.object(mysql.MySQLConfigHelper, 'get_mem_total')
    @mock.patch.object(mysql, 'log')
    def test_parse_config_innodb_invalid_values(self, mog, mem):
        mem.return_value = "512M"
        self.config_data = {
            'innodb-change-buffering': 'invalid',
        }

        helper = mysql.PerconaClusterHelper()
        mysql_config = helper.parse_config()

        self.assertTrue('innodb_change_buffering' not in mysql_config)
        self.assertTrue('innodb_io_capacity' not in mysql_config)


class Mysql8Tests(unittest.TestCase):

    def setUp(self):
        super(Mysql8Tests, self).setUp()
        self.template = "/tmp/mysql-passwd.txt"
        self.connection = mock.MagicMock()
        self.cursor = mock.MagicMock()
        self.connection.cursor.return_value = self.cursor
        self.helper = mysql.MySQL8Helper(
            rpasswdf_template=self.template,
            upasswdf_template=self.template)
        self.helper.connection = self.connection
        self.user = "user"
        self.host = "10.5.0.21"
        self.password = "passwd"
        self.db = "mydb"

    def test_grant_exists(self):
        # With backticks
        self.cursor.fetchall.return_value = (
            ("GRANT USAGE ON *.* TO `{}`@`{}`".format(self.user, self.host),),
            ("GRANT ALL PRIVILEGES ON `{}`.* TO `{}`@`{}`"
             .format(self.db, self.user, self.host),))
        self.assertTrue(self.helper.grant_exists(self.db, self.user, self.host))

        self.cursor.execute.assert_called_with(
            "SHOW GRANTS FOR '{}'@'{}'".format(self.user, self.host))

        # With single quotes
        self.cursor.fetchall.return_value = (
            ("GRANT USAGE ON *.* TO '{}'@'{}'".format(self.user, self.host),),
            ("GRANT ALL PRIVILEGES ON '{}'.* TO '{}'@'{}'"
             .format(self.db, self.user, self.host),))
        self.assertTrue(self.helper.grant_exists(self.db, self.user, self.host))

        # Grant not there
        self.cursor.fetchall.return_value = (
            ("GRANT USAGE ON *.* TO '{}'@'{}'".format("someuser", "notmyhost"),),
            ("GRANT ALL PRIVILEGES ON '{}'.* TO '{}'@'{}'"
             .format("somedb", "someuser", "notmyhost"),))
        self.assertFalse(self.helper.grant_exists(self.db, self.user, self.host))

    def test_create_grant(self):
        self.helper.grant_exists = mock.MagicMock(return_value=False)
        self.helper.create_user = mock.MagicMock()

        self.helper.create_grant(self.db, self.user, self.host, self.password)
        self.cursor.execute.assert_called_with(
            "GRANT ALL PRIVILEGES ON `{}`.* TO '{}'@'{}'"
            .format(self.db, self.user, self.host))
        self.helper.create_user.assert_called_with(self.user, self.host, self.password)

    def test_create_user(self):
        self.helper.create_user(self.user, self.host, self.password)
        self.cursor.execute.assert_called_with(
            "CREATE USER '{}'@'{}' IDENTIFIED BY '{}'".
            format(self.user, self.host, self.password))

    def test_create_router_grant(self):
        self.helper.create_user = mock.MagicMock()

        self.helper.create_router_grant(self.user, self.host, self.password)
        _calls = [
            mock.call("GRANT CREATE USER ON *.* TO '{}'@'{}' WITH GRANT OPTION"
                      .format(self.user, self.host)),
            mock.call("GRANT SELECT, INSERT, UPDATE, DELETE, EXECUTE ON "
                      "mysql_innodb_cluster_metadata.* TO '{}'@'{}'"
                      .format(self.user, self.host)),
            mock.call("GRANT SELECT ON mysql.user TO '{}'@'{}'"
                      .format(self.user, self.host)),
            mock.call("GRANT SELECT ON "
                      "performance_schema.replication_group_members TO "
                      "'{}'@'{}'".format(self.user, self.host)),
            mock.call("GRANT SELECT ON "
                      "performance_schema.replication_group_member_stats TO "
                      "'{}'@'{}'".format(self.user, self.host)),
            mock.call("GRANT SELECT ON "
                      "performance_schema.global_variables TO "
                      "'{}'@'{}'".format(self.user, self.host))]

        self.cursor.execute.assert_has_calls(_calls)
        self.helper.create_user.assert_called_with(self.user, self.host, self.password)

    def test_configure_router(self):
        self.helper.create_user = mock.MagicMock()
        self.helper.create_router_grant = mock.MagicMock()
        self.helper.normalize_address = mock.MagicMock(return_value=self.host)
        self.helper.get_mysql_password = mock.MagicMock(return_value=self.password)

        self.assertEqual(self.password, self.helper.configure_router(self.host, self.user))
        self.helper.create_user.assert_called_with(self.user, self.host, self.password)
        self.helper.create_router_grant.assert_called_with(self.user, self.host, self.password)


class MysqlHelperTests(unittest.TestCase):

    def setUp(self):
        super(MysqlHelperTests, self).setUp()

    def test_get_prefix(self):
        _tests = {
            "prefix1": "prefix1_username",
            "prefix2": "prefix2_database",
            "prefix3": "prefix3_hostname"}

        for key in _tests.keys():
            self.assertEqual(
                key,
                mysql.get_prefix(_tests[key]))

    def test_get_db_data(self):
        _unprefixed = "myprefix"
        # Test relation data has every variation of shared-db/db-router data
        _relation_data = {
            "egress-subnets": "10.5.0.43/32",
            "ingress-address": "10.5.0.43",
            "nova_database": "nova",
            "nova_hostname": "10.5.0.43",
            "nova_username": "nova",
            "novaapi_database": "nova_api",
            "novaapi_hostname": "10.5.0.43",
            "novaapi_username": "nova",
            "novacell0_database": "nova_cell0",
            "novacell0_hostname": "10.5.0.43",
            "novacell0_username": "nova",
            "private-address": "10.5.0.43",
            "database": "keystone",
            "username": "keystone",
            "hostname": "10.5.0.43",
            "mysqlrouter_username":
            "mysqlrouteruser",
            "mysqlrouter_hostname": "10.5.0.43"}

        _expected_data = OrderedDict([
            ('nova', OrderedDict([('database', 'nova'),
                                  ('hostname', '10.5.0.43'),
                                  ('username', 'nova')])),
            ('novaapi', OrderedDict([('database', 'nova_api'),
                                     ('hostname', '10.5.0.43'),
                                     ('username', 'nova')])),
            ('novacell0', OrderedDict([('database', 'nova_cell0'),
                                       ('hostname', '10.5.0.43'),
                                       ('username', 'nova')])),
            ('mysqlrouter', OrderedDict([('username', 'mysqlrouteruser'),
                                         ('hostname', '10.5.0.43')])),
            ('myprefix', OrderedDict([('hostname', '10.5.0.43'),
                                      ('database', 'keystone'),
                                      ('username', 'keystone')]))])

        _results = mysql.get_db_data(_relation_data, unprefixed=_unprefixed)

        for prefix in _expected_data.keys():
            for key in _expected_data[prefix].keys():
                self.assertEqual(
                    _results[prefix][key], _expected_data[prefix][key])



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2013 Canonical Ltd.
#
# Authors:
#  Charm Helpers Developers <juju@lists.ubuntu.com>
"""A helper to create a yaml cache of config with namespaced relation data."""
import os
import yaml

import six

import charmhelpers.core.hookenv


charm_dir = os.environ.get('CHARM_DIR', '')


def dict_keys_without_hyphens(a_dict):
    """Return the a new dict with underscores instead of hyphens in keys."""
    return dict(
        (key.replace('-', '_'), val) for key, val in a_dict.items())


def update_relations(context, namespace_separator=':'):
    """Update the context with the relation data."""
    # Add any relation data prefixed with the relation type.
    relation_type = charmhelpers.core.hookenv.relation_type()
    relations = []
    context['current_relation'] = {}
    if relation_type is not None:
        relation_data = charmhelpers.core.hookenv.relation_get()
        context['current_relation'] = relation_data
        # Deprecated: the following use of relation data as keys
        # directly in the context will be removed.
        relation_data = dict(
            ("{relation_type}{namespace_separator}{key}".format(
                relation_type=relation_type,
                key=key,
                namespace_separator=namespace_separator), val)
            for key, val in relation_data.items())
        relation_data = dict_keys_without_hyphens(relation_data)
        context.update(relation_data)
        relations = charmhelpers.core.hookenv.relations_of_type(relation_type)
        relations = [dict_keys_without_hyphens(rel) for rel in relations]

    context['relations_full'] = charmhelpers.core.hookenv.relations()

    # the hookenv.relations() data structure is effectively unusable in
    # templates and other contexts when trying to access relation data other
    # than the current relation. So provide a more useful structure that works
    # with any hook.
    local_unit = charmhelpers.core.hookenv.local_unit()
    relations = {}
    for rname, rids in context['relations_full'].items():
        relations[rname] = []
        for rid, rdata in rids.items():
            data = rdata.copy()
            if local_unit in rdata:
                data.pop(local_unit)
            for unit_name, rel_data in data.items():
                new_data = {'__relid__': rid, '__unit__': unit_name}
                new_data.update(rel_data)
                relations[rname].append(new_data)
    context['relations'] = relations


def juju_state_to_yaml(yaml_path, namespace_separator=':',
                       allow_hyphens_in_keys=True, mode=None):
    """Update the juju config and state in a yaml file.

    This includes any current relation-get data, and the charm
    directory.

    This function was created for the ansible and saltstack
    support, as those libraries can use a yaml file to supply
    context to templates, but it may be useful generally to
    create and update an on-disk cache of all the config, including
    previous relation data.

    By default, hyphens are allowed in keys as this is supported
    by yaml, but for tools like ansible, hyphens are not valid [1].

    [1] http://www.ansibleworks.com/docs/playbooks_variables.html#what-makes-a-valid-variable-name
    """
    config = charmhelpers.core.hookenv.config()

    # Add the charm_dir which we will need to refer to charm
    # file resources etc.
    config['charm_dir'] = charm_dir
    config['local_unit'] = charmhelpers.core.hookenv.local_unit()
    config['unit_private_address'] = charmhelpers.core.hookenv.unit_private_ip()
    config['unit_public_address'] = charmhelpers.core.hookenv.unit_get(
        'public-address'
    )

    # Don't use non-standard tags for unicode which will not
    # work when salt uses yaml.load_safe.
    yaml.add_representer(six.text_type,
                         lambda dumper, value: dumper.represent_scalar(
                             six.u('tag:yaml.org,2002:str'), value))

    yaml_dir = os.path.dirname(yaml_path)
    if not os.path.exists(yaml_dir):
        os.makedirs(yaml_dir)

    if os.path.exists(yaml_path):
        with open(yaml_path, "r") as existing_vars_file:
            existing_vars = yaml.load(existing_vars_file.read())
    else:
        with open(yaml_path, "w+"):
            pass
        existing_vars = {}

    if mode is not None:
        os.chmod(yaml_path, mode)

    if not allow_hyphens_in_keys:
        config = dict_keys_without_hyphens(config)
    existing_vars.update(config)

    update_relations(existing_vars, namespace_separator)

    with open(yaml_path, "w+") as fp:
        fp.write(yaml.dump(existing_vars, default_flow_style=False))



import shutil
import tempfile
import yaml

import tests.utils

import charmhelpers.contrib.openstack.files.policy_rc_d_script as policy_rc_d


NOVS_POLICY_FILE_CONTENTS = """
blocked_actions:
  neutron-dhcp-agent: [restart, stop, try-restart]
  neutron-l3-agent: [restart, stop, try-restart]
policy_requestor_name: neutron-openvswitch
policy_requestor_type: charm
"""

COMPUTE_POLICY_FILE_CONTENTS = """
blocked_actions:
  libvirt: [restart, stop, try-restart]
  nova-compute: [restart, stop, try-restart]
policy_requestor_name: nova-compute
policy_requestor_type: charm
"""

ANOTHER_POLICY_FILE_CONTENTS = """
blocked_actions:
  libvirt: [restart, stop, try-restart]
policy_requestor_name: mycharm
policy_requestor_type: charm
"""


class PolicyRCDScriptTestCase(tests.utils.BaseTestCase):

    def setUp(self):
        super(PolicyRCDScriptTestCase, self).setUp()
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        super(PolicyRCDScriptTestCase, self).tearDown()
        shutil.rmtree(self.test_dir)

    def write_policy(self, pfile_name, contents):
        test_policy_file = '{}/{}'.format(self.test_dir, pfile_name)
        with open(test_policy_file, 'w') as f:
            f.write(contents)
        return test_policy_file

    def test_read_policy_file(self):
        test_policy_file = self.write_policy(
            'novs.policy',
            NOVS_POLICY_FILE_CONTENTS)
        policy = policy_rc_d.read_policy_file(test_policy_file)
        self.assertEqual(
            policy,
            [
                policy_rc_d.SystemPolicy(
                    policy_requestor_name='neutron-openvswitch',
                    policy_requestor_type='charm',
                    service='neutron-dhcp-agent',
                    blocked_actions=['restart', 'stop', 'try-restart']),
                policy_rc_d.SystemPolicy(
                    policy_requestor_name='neutron-openvswitch',
                    policy_requestor_type='charm',
                    service='neutron-l3-agent',
                    blocked_actions=['restart', 'stop', 'try-restart'])])

    def test_get_policies(self):
        self.write_policy('novs.policy', NOVS_POLICY_FILE_CONTENTS)
        self.write_policy('compute.policy', COMPUTE_POLICY_FILE_CONTENTS)
        policies = policy_rc_d.get_policies(self.test_dir)
        self.maxDiff = None
        self.assertEqual(
            sorted(policies),
            [
                policy_rc_d.SystemPolicy(
                    policy_requestor_name='neutron-openvswitch',
                    policy_requestor_type='charm',
                    service='neutron-dhcp-agent',
                    blocked_actions=['restart', 'stop', 'try-restart']),
                policy_rc_d.SystemPolicy(
                    policy_requestor_name='neutron-openvswitch',
                    policy_requestor_type='charm',
                    service='neutron-l3-agent',
                    blocked_actions=['restart', 'stop', 'try-restart']),
                policy_rc_d.SystemPolicy(
                    policy_requestor_name='nova-compute',
                    policy_requestor_type='charm',
                    service='libvirt',
                    blocked_actions=['restart', 'stop', 'try-restart']),
                policy_rc_d.SystemPolicy(
                    policy_requestor_name='nova-compute',
                    policy_requestor_type='charm',
                    service='nova-compute',
                    blocked_actions=['restart', 'stop', 'try-restart'])])

    def test_record_blocked_action(self):
        self.patch_object(policy_rc_d.time, 'time')
        self.time.return_value = 456
        self.patch_object(policy_rc_d.uuid, 'uuid1')
        uuids = ['uuid1', 'uuid2']
        self.uuid1.side_effect = lambda: uuids.pop()
        blocking_policies = [
            policy_rc_d.SystemPolicy(
                policy_requestor_name='cinder',
                policy_requestor_type='charm',
                service='cinder',
                blocked_actions=['restart', 'stop', 'try-restart']),
            policy_rc_d.SystemPolicy(
                policy_requestor_name='cinder-ceph',
                policy_requestor_type='charm',
                service='cinder',
                blocked_actions=['restart', 'stop', 'try-restart'])]
        policy_rc_d.record_blocked_action(
            'cinder-api',
            'restart',
            blocking_policies,
            self.test_dir)
        expect = [
            (
                '{}/charm-cinder-uuid2.deferred'.format(self.test_dir),
                {
                    'action': 'restart',
                    'reason': 'Package update',
                    'policy_requestor_name': 'cinder',
                    'policy_requestor_type': 'charm',
                    'service': 'cinder-api',
                    'timestamp': 456.0}),
            (
                '{}/charm-cinder-ceph-uuid1.deferred'.format(self.test_dir),
                {
                    'action': 'restart',
                    'reason': 'Package update',
                    'policy_requestor_name': 'cinder-ceph',
                    'policy_requestor_type': 'charm',
                    'service': 'cinder-api',
                    'timestamp': 456.0})]
        for defer_file, contents in expect:
            with open(defer_file, 'r') as f:
                self.assertEqual(
                    yaml.safe_load(f),
                    contents)

    def test_get_blocking_policies(self):
        self.write_policy('novs.policy', NOVS_POLICY_FILE_CONTENTS)
        self.write_policy('compute.policy', COMPUTE_POLICY_FILE_CONTENTS)
        policies = policy_rc_d.get_blocking_policies(
            'libvirt',
            'restart',
            self.test_dir)
        self.assertEqual(
            policies,
            [
                policy_rc_d.SystemPolicy(
                    policy_requestor_name='nova-compute',
                    policy_requestor_type='charm',
                    service='libvirt',
                    blocked_actions=['restart', 'stop', 'try-restart'])])

    def test_process_action_request(self):
        self.write_policy('novs.policy', NOVS_POLICY_FILE_CONTENTS)
        self.write_policy('compute.policy', COMPUTE_POLICY_FILE_CONTENTS)
        self.write_policy('another.policy', ANOTHER_POLICY_FILE_CONTENTS)
        self.assertEqual(
            policy_rc_d.process_action_request(
                'libvirt',
                'restart',
                self.test_dir,
                self.test_dir),
            (
                False,
                ('restart of libvirt blocked by charm mycharm, '
                 'charm nova-compute')))
        self.assertEqual(
            policy_rc_d.process_action_request(
                'glance-api',
                'restart',
                self.test_dir,
                self.test_dir),
            (
                True,
                'Permitting glance-api restart'))



#!/usr/bin/env python

from charmhelpers.contrib.sysctl.watermark_scale_factor import (
    watermark_scale_factor,
    calculate_watermark_scale_factor,
    get_normal_managed_pages,
)

from mock import patch
import unittest

from tests.helpers import patch_open

TO_PATCH = [
    "log",
    "ERROR",
    "DEBUG"
]

PROC_ZONEINFO = """
Node 0, zone   Normal
  pages free     1253032
        min      16588
        low      40833
        high     65078
        spanned  24674304
        present  24674304
        managed  24247810
        protection: (0, 0, 0, 0, 0)
"""


class TestWatermarkScaleFactor(unittest.TestCase):

    def setUp(self):
        for m in TO_PATCH:
            setattr(self, m, self._patch(m))

    def _patch(self, method):
        _m = patch('charmhelpers.contrib.sysctl.watermark_scale_factor.' + method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        return mock

    @patch('charmhelpers.contrib.sysctl.watermark_scale_factor.get_normal_managed_pages')
    @patch('charmhelpers.core.host.get_total_ram')
    def test_calculate_watermark_scale_factor(self, get_total_ram, get_normal_managed_pages):
        get_total_ram.return_value = 101254156288
        get_normal_managed_pages.return_value = [24247810]
        wmark = calculate_watermark_scale_factor()
        self.assertTrue(wmark >= 10, "ret {}".format(wmark))
        self.assertTrue(wmark <= 1000, "ret {}".format(wmark))

    def test_get_normal_managed_pages(self):
        with patch_open() as (mock_open, mock_file):
            mock_file.readlines.return_value = PROC_ZONEINFO.splitlines()
            self.assertEqual(get_normal_managed_pages(), [24247810])
            mock_open.assert_called_with('/proc/zoneinfo', 'r')

    def test_watermark_scale_factor(self):
        mem_totals = [17179803648, 34359607296, 549753716736]
        managed_pages = [4194288, 24247815, 8388576, 134217216]
        arglists = [[mem, managed] for mem in mem_totals for managed in managed_pages]

        for arglist in arglists:
            wmark = watermark_scale_factor(*arglist)
            self.assertTrue(wmark >= 10, "assert failed for args: {}, ret {}".format(arglist, wmark))
            self.assertTrue(wmark <= 1000, "assert failed for args: {}, ret {}".format(arglist, wmark))



#!/usr/bin/env python
# coding: utf-8

from unittest import TestCase

from charmhelpers.fetch.python import debug

import mock

__author__ = "Jorge Niedbalski <jorge.niedbalski@canonical.com>"

TO_PATCH = [
    "log",
    "open_port",
    "close_port",
    "Rpdb",
    "_error",
]


class DebugTestCase(TestCase):
    """Test cases for charmhelpers.contrib.python.debug"""

    def setUp(self):
        TestCase.setUp(self)
        self.patch_all()
        self.log.return_value = True
        self.close_port.return_value = True

        self.wrapped_function = mock.Mock(return_value=True)
        self.set_trace = debug.set_trace

    def patch(self, method):
        _m = mock.patch.object(debug, method)
        _mock = _m.start()
        self.addCleanup(_m.stop)
        return _mock

    def patch_all(self):
        for method in TO_PATCH:
            setattr(self, method, self.patch(method))

    def test_debug_set_trace(self):
        """Check if set_trace works
        """
        self.set_trace()
        self.open_port.assert_called_with(4444)

    def test_debug_set_trace_ex(self):
        """Check if set_trace raises exception
        """
        self.set_trace()
        self.Rpdb.set_trace.side_effect = Exception()
        self.assertTrue(self._error.called)
